/*****************************************************************************
**
** null.c
*/
static char st_null_version[] = "$Header: /home/omer/work/cvs_repository/Hebrew/include/null.c,v 1.12 1999/02/06 18:22:11 omer Exp $";
/*
** Level 0 null text rendering plugin
** Was developed in order to test the level 0 interface specifications.
** This plugin does no reordering or transformation of rendered characters.
** It merely copies characters from the original string to the rendered
** output.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

/* #define DEBUG */

#include <assert.h>
#include <string.h>

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

#include "../include/plugin0.h"

/*****************************************************************************
** Dynamic library initialization and finalization functions                **
*****************************************************************************/
void
_init(void)
{
#ifdef DEBUG
  fprintf(stderr,"_init() was called.\n");
#endif /* DEBUG */
}

void
_fini(void)
{
#ifdef DEBUG
  fprintf(stderr,"_fini() was called.\n");
#endif /* DEBUG */
}

/*****************************************************************************
** The following is the main connection point between the plugin
** and the client software.
*****************************************************************************/

SRenderingFunctions st_null_fp = {
  sizeof(SRenderingFunctions),      /* size_t m_struct_length; */
  LEVEL0_VERSION_STRING,            /* char m_preamble_str[9]; */
  &parse_config,                    /* void *m_parse_config_func; */
  &delete_config,                   /* void *m_delete_config_func; */
  &p_notify_modification,           /* void *m_notify_modification_func; */
  &p_retrieve_rendered_line,        /* void *m_retrieve_rendered_line_func; */
  LEVEL0_VERSION_STRING             /* char m_posteamble_str[9]; */
};

const SRenderingFunctions *
return_function_pointers(void)
{
  return(&st_null_fp);
}


/*****************************************************************************
** Plugin configuration functions - null.c has no configuration options
** whatsoever.
** So the configuration string must be NULL, empty or consist of only
** spaces.
*****************************************************************************/
E_RENDER_RETURN_CODES
parse_config(const char *in_config_str,
	                           /* Configuration string - typically given
				   ** as a command line argument or retrieved
				   ** from a resource file. */
	     MALLOC *in_malloc_func,
	     FREE *in_free_func,
	     void **out_plugin_config_p_p)
                                   /* Pointer to opaque configuration data
				   ** structure, which will be used by
				   ** other procedures.  If necessary,
				   ** the data structure will contain also
				   ** a pointer to a procedure to free
				   ** allocated memory (initialized
				   ** from in_free_func). */
{
  size_t l_spaces_span;

  assert(out_plugin_config_p_p != NULL);
  assert(*out_plugin_config_p_p == NULL);  /* Will not be modified here */

  if (in_config_str == NULL) return(RENDER_SUCCESS);
  if (strlen(in_config_str) == 0) return(RENDER_SUCCESS);

  l_spaces_span = strspn(in_config_str," ");
  if (strlen(in_config_str) == l_spaces_span) {
    return(RENDER_SUCCESS);
  }
  else {
    return(RENDER_FAILURE_INVALID_ARGUMENTS);
  }
}

BOOL
validate_config(const void *in_plugin_config_p)
                                   /* Pointer to plugin configuration opaque
				   ** data structure, whose contents are to
				   ** be validated. */
{
  /* The only valid configuration for this plugin is the
  ** NULL configuration. */
  return(NULL == in_plugin_config_p);
}

void
delete_config(void *in_plugin_config_p)
{
  assert(NULL == in_plugin_config_p);
}


/*****************************************************************************
** Notify the plugin about a modification in the original text.
** The client-accessible function encapsulates a call to the plugin-provided
** function called p_notify_modification() and with the same arguments.
**
** The basic idea is that a change consists of deletion of a text segment
** starting at offset N and having length L1, followed by insertion
** of another text segment starting at the same offset and having length L2.
** Either text segment may have zero length.
** Thus, the same mechanism is used to describe both text replacements,
** insertions and deletions.
*****************************************************************************/

E_RENDER_RETURN_CODES
p_notify_modification(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
		      size_t in_change_is_at,
		                  /* The offset in the original text
                                  ** in which the change took place. */
		      size_t in_deleted_chars,
		                  /* Number of deleted characters */
		      size_t in_inserted_chars,
		                  /* Number of inserted characters */

		      GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
		      void *in_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the callback. */

		      size_t *out_first_modified_line_p)
		                  /* Reply to the client software - which is
				  ** the first rendered line to be affected
				  ** by the change. */
{
  size_t l_line_index;
  E_RENDER_RETURN_CODES ret;

  assert(NULL != in_out_text_p);
  assert(NULL != out_first_modified_line_p);

  ret = locate_fragment(in_out_text_p, in_change_is_at,
			&l_line_index, NULL,
			NULL, NULL);
  if (RENDER_FAILURE_NOT_RENDERED_YET == ret) {
    /* If the modified line is a fragment, which does not exist yet,
    ** then we want to return the index of the first line not rendered
    ** yet. */

    *out_first_modified_line_p = in_out_text_p->m_used_line_array_size;

    /* In this case, the text_delete_lines() will do nothing, but
    ** it has an assert statement against the present case,
    ** so we'll be good boys and skip it. */

#ifdef DEBUG
    fprintf(stderr,"p_notify_modification(at=%u,del=%u,ins=%u) line beyond last rendered line=%u\n",
	    (unsigned int) in_change_is_at,
	    (unsigned int) in_deleted_chars,
	    (unsigned int) in_inserted_chars,
	    (unsigned int) (*out_first_modified_line_p));
#endif /* DEBUG */

    return(RENDER_SUCCESS);
  }
  else if (RENDER_SUCCESS != ret) return(ret);

  *out_first_modified_line_p = l_line_index;

#ifdef DEBUG
  fprintf(stderr,"p_notify_modification(at=%u,del=%u,ins=%u) line=%u\n",
	  (unsigned int) in_change_is_at,
	  (unsigned int) in_deleted_chars,
	  (unsigned int) in_inserted_chars,
	  (unsigned int) l_line_index);
#endif /* DEBUG */

  /* We delete all lines starting from the modified line.
  ** We delete even fragments which precede the modification point. */
  ret = text_delete_lines(in_out_text_p, l_line_index);
  return(ret);
}


/*****************************************************************************
** Retrieve a line of rendered glyphs.
** Most plugin writers will probably implement the entire rendering
** algorithm in this procedure, because this is the right place for
** lazy execution of the rendering algorithm.
*****************************************************************************/

E_RENDER_RETURN_CODES
p_retrieve_rendered_line(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
			 size_t in_desired_visual_row,
			          /* Index of the line which the client
				  ** software wants to retrieve. */
			 GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
			 void *in_frag_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_frag_func
				  ** callback. */
			 GET_RENDERED_LINE_SIZE *in_get_line_size_func,
			          /* Client-provided callback for
				  ** retrieving desired rendered line
				  ** size. */
			 void *in_size_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_line_size_func
				  ** callback. */
			 /* Output arguments - returned only if
			 ** out_glyphs_arr_p is not NULL. */
			 size_t *out_line_length_p,
			         /* The array length is in glyphs
				 ** (which could be multi-byte). */
			 void **out_glyphs_arr_p)
{
  E_RENDER_RETURN_CODES ret;
  size_t l_tmp;
  size_t l_current_visual_row;
  BOOL l_end_of_text_flag = FALSE;

  assert(NULL != in_out_text_p);
  /* Here, we are not validating inside the SRenderedText structure. */
  assert(NULL != in_get_frag_func);
  assert(NULL != in_get_line_size_func);
  assert((NULL == out_glyphs_arr_p) || (NULL != out_line_length_p));
  assert((NULL == out_glyphs_arr_p) || (NULL == *out_glyphs_arr_p));

#ifdef DEBUG
  fprintf(stderr,"p_retrieve_rendered_line(want row %u) was called.\n",
	  (unsigned int) in_desired_visual_row);
#endif /* DEBUG */

  /* If the requested line is not rendered yet, then render all
  ** lines after the last rendered line until the requested line
  ** (inclusive). */

  l_tmp = in_out_text_p->m_used_line_array_size;  /* m_used_line_array_size
						  ** will be modified in the
						  ** loop */
  for (l_current_visual_row = l_tmp;
       l_current_visual_row <= in_desired_visual_row;
       l_current_visual_row++) {
    size_t l_orig_text_offset;    /* Running offset into the original
				  ** text, starting from the first byte
				  ** of the original text to be rendered
				  ** into the current line. */
    size_t l_orig_frag_length;    /* Length of the current fragment
				  ** from the original text. */
    long l_orig_frag_attributes;
    char *l_orig_frag_text_p;
    size_t l_orig_frag_index;     /* Running index to the original fragment.
				  ** Its limit is l_orig_frag_length. */
    size_t l_remaining_visual_line_size;
                                  /* How many glyphs remain to be copied
				  ** to the rendered line until it is
				  ** finished.
				  ** In the null plugin, this is redundant
				  ** because there is 1:1 correspondence
				  ** between original text and rendered text
				  ** characters.  But I wanted to develop
				  ** general code, which can be reused also
				  ** in other plugins with minimum
				  ** modifications. */
    size_t l_latched_remaining_visual_line_size;
                                  /* At the start of each fragment,
				  ** l_remaining_visual_line_size is
				  ** latched into this variable. */
    size_t l_rendered_frag_index; /* Index of next glyph in the
				  ** rendered fragment. */

    SRenderedLine *l_line_p;
    SRenderedFragment *l_frag_p = NULL;  /* NULL value indicates that there
				  ** is no previous rendered text fragment
				  ** to be closed. */

#ifdef DEBUG
    fprintf(stderr,"p_retrieve_rendered_line - going to render row %u.\n",
	    (unsigned int) l_current_visual_row);
#endif /* DEBUG */

    ret = locate_origtext_offset(in_out_text_p,
				 l_current_visual_row,
				 &l_orig_text_offset);
    if (RENDER_SUCCESS != ret) return(ret);
    ret = (*in_get_line_size_func)(in_size_client_data_p,
				   l_current_visual_row,
				   &l_remaining_visual_line_size);
    if (RENDER_SUCCESS != ret) return(ret);

#ifdef DEBUG
    fprintf(stderr,"p_retrieve_rendered_line - orig. text offset=%u, visual line size=%u.\n",
	    (unsigned int) l_orig_text_offset,
	    (unsigned int) l_remaining_visual_line_size);
#endif /* DEBUG */

    l_line_p = construct_line(in_out_text_p->m_rendering_context_p,
			      l_remaining_visual_line_size);
    if (NULL == l_line_p) return(RENDER_FAILURE_NO_MEMORY);

    /* Trigger request to fetch 1st original text fragment */
    l_orig_frag_length = 0;
    l_orig_frag_index = 0;
    while(l_remaining_visual_line_size > 0) {
      if (l_orig_frag_index >= l_orig_frag_length) {
	/* Need to retrieve another original text fragment, as well
	** as moving to the next rendered text fragment. */

	/* Close previous rendered text fragment, if any. */
	if (NULL != l_frag_p) {
	  ret = line_append_fragment(in_out_text_p->m_rendering_context_p,
				l_line_p,
				l_frag_p);
	  if (RENDER_SUCCESS != ret) return(ret);
	  l_frag_p = NULL;
	}

	/* Actually retrieve original text fragment. */
	l_orig_frag_text_p = NULL;
	ret = (*in_get_frag_func)(in_frag_client_data_p,
				  l_orig_text_offset,
				  &l_orig_frag_length,
				  &l_orig_frag_attributes,
				  &l_orig_frag_text_p);
	if (RENDER_BEYOND_END_OF_TEXT == ret) {
	  l_end_of_text_flag = TRUE;
#ifdef DEBUG
	  fprintf(stderr,"p_retrieve_rendered_line - reached end of"
		  " original text\n");
#endif /* DEBUG */
	  break;
	}
	else if (RENDER_SUCCESS != ret) return(ret);
	l_orig_frag_index = 0;

	/* Open a new rendered text fragment. */
	l_latched_remaining_visual_line_size = l_remaining_visual_line_size;
	assert(NULL == l_frag_p);
	l_frag_p = construct_fragment(in_out_text_p->m_rendering_context_p,
				      min(l_latched_remaining_visual_line_size,
					  l_orig_frag_length),
				      l_orig_frag_attributes,
				      0,  /* No added frag attributes */
				      l_orig_text_offset,
				      min(l_latched_remaining_visual_line_size,
					  l_orig_frag_length));
	if (NULL == l_frag_p) return(RENDER_FAILURE_NO_MEMORY);
	/* Initialize fields not initialized by construct_fragment(). */
	l_frag_p->m_frag_state_free_func = NULL;
	l_frag_p->m_frag_state_p = NULL;

	l_rendered_frag_index = 0;
      }

#ifdef DEBUG_EXTRA
      fprintf(stderr,"p_retrieve_rendered_line - l_rendered_frag_index=%u, "
	      "l_latched_remaining_visual_line_size=%u, "
	      "l_orig_frag_length=%u.\n",
	      (unsigned int) l_rendered_frag_index,
	      (unsigned int) l_latched_remaining_visual_line_size,
	      (unsigned int) l_orig_frag_length);
#endif /* DEBUG_EXTRA */

      assert(l_rendered_frag_index < min(l_latched_remaining_visual_line_size,
					 l_orig_frag_length));

      ((char *)(l_frag_p->m_frag_str))[l_rendered_frag_index]
	= *l_orig_frag_text_p;
      l_frag_p->m_back_xrefs_arr[l_rendered_frag_index] = l_orig_text_offset;
      l_frag_p->m_forward_xrefs_arr[l_orig_frag_index] = l_rendered_frag_index;

      /* Increment indexes, pointers, etc. */
      l_orig_text_offset++;
      l_orig_frag_text_p++;
      l_orig_frag_index++;
      l_rendered_frag_index++;
      l_remaining_visual_line_size--;
    }
    /* Close last rendered text fragment. */
    if (NULL != l_frag_p) {
      ret = line_append_fragment(in_out_text_p->m_rendering_context_p,
				 l_line_p,
				 l_frag_p);
      if (RENDER_SUCCESS != ret) return(ret);
    }
    else {
      /* If we hit end of text file before creating any fragments, then
      ** clean up and return an appropriate return code to the caller. */
      if (FALSE != l_end_of_text_flag) {
	if (0 == l_line_p->m_used_frag_array_size) {
	  destroy_line(l_line_p);
	  return(RENDER_BEYOND_END_OF_TEXT);
	}
      }
    }
    /* Now append the newly-constructed line to the rendered text
    ** structure. */
    ret = text_append_line(in_out_text_p, l_line_p);
    if (RENDER_SUCCESS != ret) return(ret);
    l_line_p = NULL;
  }
  assert(in_out_text_p->m_used_line_array_size > in_desired_visual_row);

  /* Check if the caller wants a concatenated version of the line
  ** fragments.  Otherwise, the caller may have called this procedure
  ** just to ensure that the desired rendered line exists. */
  if (NULL == out_glyphs_arr_p) {
    return(RENDER_SUCCESS);
  }
  else {
    /* The caller wants a concatenated version, so build one for him.
    ** We play it lazy, so we allocate memory for the maximum line length
    ** without subsequent attempt to reallocate memory according to actual
    ** usage. */

    size_t l_frag_index;
    size_t l_frag_offset;          /* in glyphs */

    *out_glyphs_arr_p = in_out_text_p->m_rendering_context_p->
      m_malloc_func(in_out_text_p->m_rendering_context_p->m_bytes_per_glyph
		    * in_out_text_p->m_line_p_arr[in_desired_visual_row]
		                                     ->m_screen_row_width);
    if (NULL == *out_glyphs_arr_p) {
      return(RENDER_FAILURE_NO_MEMORY);
    }

    assert(in_out_text_p->m_line_p_arr[in_desired_visual_row]
	   ->m_used_frag_array_size > 0);

    for (l_frag_index = 0, l_frag_offset = 0;
	 l_frag_index <
	   in_out_text_p->m_line_p_arr[in_desired_visual_row]
	                                            ->m_used_frag_array_size;
	 l_frag_index++) {
      SRenderedFragment *l_curr_frag_p =
	in_out_text_p->m_line_p_arr[in_desired_visual_row]
	->m_frag_p_arr[l_frag_index];

      memmove((*out_glyphs_arr_p) + l_frag_offset,   /* dest */
	      l_curr_frag_p->m_frag_str,             /* src */
	      l_curr_frag_p->m_frag_length
	    * in_out_text_p->m_rendering_context_p->m_bytes_per_glyph);

      l_frag_offset += l_curr_frag_p->m_frag_length;
    }
    assert(NULL != out_line_length_p);
    *out_line_length_p = l_frag_offset;

#ifdef DEBUG
    fprintf(stderr,"p_retrieve_rendered_line (end): out line length=%u"
	    ", desired visual row=%u, row width=%u\n",
	    (unsigned int) (*out_line_length_p),
	    (unsigned int) in_desired_visual_row,
	    (unsigned int) in_out_text_p->m_line_p_arr[in_desired_visual_row]
	                                      ->m_screen_row_width);
    dump_SRenderedText(in_out_text_p, stderr);
#endif /* DEBUG */

    assert(*out_line_length_p
	   <= in_out_text_p->m_line_p_arr[in_desired_visual_row]
	                                      ->m_screen_row_width);
    return(RENDER_SUCCESS);
  }
}

/****************************************************************************/

/* End of null.c */
