/*****************************************************************************
**
** plugclient0.c
*/
static char st_plugclient0_version[] = "$Header: /home/omer/work/cvs_repository/Hebrew/include/plugclient0.c,v 1.6 1999/02/06 18:22:11 omer Exp $";
/*
** Implements some procedures, which are needed for working with all
** level 0 text rendering plugins.
** Those procedures are invoked by the client application.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

#include <stdlib.h>
#include <dlfcn.h>
#include <stdio.h>
#include <assert.h>
#include "structs0.h"
#include "plugimp0.h"
#include "plugin0.h"
#include "plugclient0.h"


/*****************************************************************************
** bind_to_plugin()
** Finds the requested plugin and binds the shared library containing the
** plugin.
** The plugin will be configured by set_rendering_context().
*****************************************************************************/

E_RENDER_RETURN_CODES
bind_to_plugin(MALLOC *in_malloc_func,
	       FREE *in_free_func,
	                           /* Functions to be used for
				   ** allocating and freeing memory
				   ** as needed by this procedure.
				   ** No rendering context exists at
				   ** this point, so it can't be used.
				   ** Default values of malloc() and
				   ** free() are used for NULL arguments. */
	       const char *in_plugin_name_str,
	                           /* Name of plugin (typically, the name
				   ** of a file in a standard directory). */
	       SPlugin *out_plugin_p)
                                   /* Output argument, which will
				   ** receive bound plugin data. */
{
  /* Local variables */
  char *l_libpath_str = NULL; /* Pointer to library path string */
  int l_plugin_name_length;
  char *l_plugin_name_str = NULL; /* Full name of the plugin */
  void *l_lib_handle;             /* Handle for the loaded shared library */
  RETURN_FUNCTION_POINTERS *l_rfp_func;
  const char *l_error;
  const SRenderingFunctions *l_funcptrs_p;
                                  /* Pointer to structure pointing at the
				  ** various functions of the plugin */

  /* Validate arguments */
  assert(NULL != in_plugin_name_str);
  assert(0 != strlen(in_plugin_name_str));
  assert(NULL != out_plugin_p);
  /* Ensure that *out_plugin_p is not already in use. */
  assert(NULL == out_plugin_p->m_lib_handle);
  assert(NULL == out_plugin_p->m_funcptrs_p);

  if (NULL == in_malloc_func) in_malloc_func = &malloc;
  if (NULL == in_free_func) in_free_func = &free;

  /* Locate and load the library */
  l_libpath_str = getenv(PLUGIN_PATH_ENVIRONMENT_VARIABLE_NAME);
  if (l_libpath_str == NULL) l_libpath_str = HARDWIRED_PLUGIN_PATH_DEFAULT;
  /* Note:  see structs0.h for definitions of the macros used here. */

  l_plugin_name_length = strlen(l_libpath_str)
    + strlen(in_plugin_name_str) + 100;
  /* The 100 are for miscellaneous characters.  I am too lazy to count
  ** exactly how many characters are needed.  100 bytes of dynamic
  ** memory are cheaper! */

  l_plugin_name_str = (char *)(*in_malloc_func)(l_plugin_name_length);
  if (l_plugin_name_str == NULL) {
    fprintf(stderr,
	    "Error in bind_to_plugin(%s) - memory allocation failure.\n",
	    in_plugin_name_str);
    return(RENDER_FAILURE_NO_MEMORY);
  }
  sprintf(l_plugin_name_str, "%s/%s.so", l_libpath_str, in_plugin_name_str);

  /* DEBUG ** fprintf(stderr,"Will load plugin %s\n",l_plugin_name_str); */

  l_lib_handle = dlopen(l_plugin_name_str, RTLD_LAZY);
  if (!l_lib_handle) {
    fprintf(stderr, "Error in opening library %s:  %s\n",
	    l_plugin_name_str, dlerror());

    (*in_free_func)(l_plugin_name_str);
    l_plugin_name_str = NULL;
    return(RENDER_FAILURE_GENERAL);
  }

  /* Retrieve pointers to functions in the library */
  l_rfp_func = dlsym(l_lib_handle, "return_function_pointers");
  l_error = dlerror();
  if (l_error != NULL) {
    fprintf(stderr,
	    "Error in retrieving return_function_pointers():  %s\n",
	    l_error);

    (*in_free_func)(l_plugin_name_str);
    l_plugin_name_str = NULL;
    return(RENDER_FAILURE_GENERAL);
  }
  l_funcptrs_p = l_rfp_func();

  if (validate_function_pointers(l_funcptrs_p) == FALSE) {
    fprintf(stderr,
	    "Plugin %s is incompatible in type or version\n",
	    l_plugin_name_str);

    (*in_free_func)(l_plugin_name_str);
    l_plugin_name_str = NULL;
    return(RENDER_FAILURE_GENERAL);
  }

  /* Return the discovered information to the client. */
  out_plugin_p->m_lib_handle = l_lib_handle;
  out_plugin_p->m_funcptrs_p = l_funcptrs_p;

  (*in_free_func)(l_plugin_name_str);
  l_plugin_name_str = NULL;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** The following function is used by bind_to_plugin() to validate the
** plugin type and version being used.
*****************************************************************************/
BOOL
validate_function_pointers(const SRenderingFunctions *in_rf_p)
{
  if (NULL == in_rf_p) return(FALSE);
  if (sizeof(SRenderingFunctions) != in_rf_p->m_struct_length) return(FALSE);
  if (strcmp(LEVEL0_VERSION_STRING,in_rf_p->m_preamble_str) != 0) {
    return(FALSE);
  }
  if (strcmp(LEVEL0_VERSION_STRING,in_rf_p->m_posteamble_str) != 0) {
    return(FALSE);
  }

  return(TRUE);
}

/*****************************************************************************
** unbind_from_plugin()
** Releases the plugin.  If no one else is using the same plugin, the
** corresponding shared library is released from memory.
*****************************************************************************/

E_RENDER_RETURN_CODES
unbind_from_plugin(SPlugin *in_out_plugin_p)
                                   /* As input, this structure identifies
				   ** the shared library handle to be freed.
				   ** When done, all fields in this structure
				   ** are set to NULL. */
{
  assert(NULL != in_out_plugin_p);
  /* We are not validating in_out_plugin_p->m_lib_handle, because it
  ** might happen to be NULL as far as we are concerned. */
  assert(NULL != in_out_plugin_p->m_funcptrs_p);

  dlclose(in_out_plugin_p->m_lib_handle);

  /* Mark this SPlugin instance as UNUSED. */
  in_out_plugin_p->m_lib_handle = NULL;
  in_out_plugin_p->m_funcptrs_p = NULL;

  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Create a new rendering context structure and set its fields.
** This includes, wherever necessary, processing of default values.
** Optionally, an existing rendering context may be supplied in order
** to copy from it values not explicitly set.
*****************************************************************************/

E_RENDER_RETURN_CODES
set_clone_rendering_context
(const SRenderingContext *in_source_rc_p,
                                   /* Original rendering context, from which
				   ** we are cloning (possibly with
				   ** modifications) the new rendering
				   ** context.
				   ** If NULL, a brand new rendering
				   ** context is to be created. */
 long int in_fields_mask,          /* OR of bitmasks defining which
				   ** fields (if any) are to be modified
				   ** from the arguments.
				   ** If RC_ALL, in_source_rc_p is
				   ** ignored. */
 SPlugin *in_plugin_p,
 const char *in_config_str,        /* Plugin configuration string
				   ** (typically, the name of a
				   ** configuration file). */
 MALLOC *in_malloc_func,
 FREE *in_free_func,
 size_t in_bytes_per_glyph,
 size_t in_initial_total_frag_array_size,
 SRenderingContext **out_context_p_p)
{
  /* Local variables */
  SRenderingContext *l_context_p = NULL;
  MALLOC *l_malloc_func = NULL;
  FREE *l_free_func = NULL;
  E_RENDER_RETURN_CODES ret;

  /* Assert statements */
  assert((NULL == in_source_rc_p)
	 || validate_rendering_context(in_source_rc_p));
  assert(NULL != in_plugin_p);
  assert(NULL != in_plugin_p->m_funcptrs_p);
  /* in_config_str can be NULL or empty. */
  assert(in_bytes_per_glyph > 0);
  assert(in_initial_total_frag_array_size > 0);
  assert(NULL != out_context_p_p);
  assert(NULL == *out_context_p_p);

  /* Select a malloc() function */
  if ((NULL == in_source_rc_p) || ((in_fields_mask & RC_MALLOC) != 0)) {
    /* Set the function from the argument list */
    l_malloc_func = in_malloc_func;
  }
  else {
    /* Set the function from the original rendering context */
    l_malloc_func = in_source_rc_p->m_malloc_func;
  }
  if (NULL == l_malloc_func) {
    l_malloc_func = &malloc;   /* Hardwired default */
  }
  /* Select a free() function */
  if ((NULL == in_source_rc_p) || ((in_fields_mask & RC_FREE) != 0)) {
    /* Set the function from the argument list */
    l_free_func = in_free_func;
  }
  else {
    /* Set the function from the original rendering context */
    l_free_func = in_source_rc_p->m_free_func;
  }
  if (NULL == l_free_func) {
    l_free_func = &free;   /* Hardwired default */
  }

  /* Allocate a rendering context */
  l_context_p = (SRenderingContext *)
    (*l_malloc_func)(sizeof(SRenderingContext));
  if (NULL == l_context_p) {
    return(RENDER_FAILURE_NO_MEMORY);
  }

  /* Set the various fields in the rendering context */
  /* 1. m_plugin_p */
  l_context_p->m_plugin_p = ((NULL == in_source_rc_p)
			     || ((in_fields_mask & RC_PLUGIN) != 0))
    ? in_plugin_p : in_source_rc_p->m_plugin_p;
  /* 2. m_plugin_config_p */
  if ((NULL == in_source_rc_p)
      || (0 != (in_fields_mask & RC_PLUGIN_CONFIG))) {
    /* Build internal configuration data structure from the
    ** configuration string in the argument list. */
    void *l_plugin_config_p = NULL;
                              /* Pointer to opaque plugin configuration
			      ** data structure */
    ret = (*((PARSE_CONFIG *)
	     (l_context_p->m_plugin_p->m_funcptrs_p->m_parse_config_func)))
      (in_config_str, l_malloc_func, l_free_func, &l_plugin_config_p);
    if (RENDER_SUCCESS != ret) {
      assert(NULL == l_plugin_config_p);
      /* We are asserting that m_parse_cohnfig_func cleans up after
      ** itself in case of failure.  So we need to free only the rendering
      ** context itself. */
      (*l_free_func)(l_context_p);
      return(ret);
    }
    l_context_p->m_plugin_config_p = l_plugin_config_p;
  } else {
    /* Set the function from the original rendering context.
    ** !!! NOTE:  we really need here a procedure for cloning
    ** !!! the configuration data structure, because it will
    ** !!! be automatically destroyed when the rendering context
    ** !!! structure is freed.
    ** !!! Until such cloning procedure is implemented, we are putting
    ** !!! here an assert(FALSE). */
    l_context_p->m_plugin_config_p = in_source_rc_p->m_plugin_config_p;
    assert(FALSE);
  }
  /* 3. m_malloc_func */
  l_context_p->m_malloc_func = l_malloc_func;
  /* 4. m_free_func */
  l_context_p->m_free_func = l_free_func;
  /* 5. m_bytes_per_glyph */
  l_context_p->m_bytes_per_glyph = ((NULL == in_source_rc_p)
			     || (0 != (in_fields_mask & RC_GLYPH_SIZE)))
    ? in_bytes_per_glyph : in_source_rc_p->m_bytes_per_glyph;
  /* 6. m_initial_total_frag_array_size */
  l_context_p-> m_initial_total_frag_array_size =
    ((NULL == in_source_rc_p) || (0 != (in_fields_mask & RC_FRAG_PER_LINE)))
    ? in_initial_total_frag_array_size
    : in_source_rc_p->m_initial_total_frag_array_size;

  /* Export pointer to the newly-created rendering context. */
  *out_context_p_p = l_context_p;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Free memory allocated to the fields of a rendering context structure.
*****************************************************************************/
void destroy_rendering_context(void *in_rendering_context_p)
{
  SRenderingContext *l_rendering_context_p =
    (SRenderingContext *) in_rendering_context_p;
  FREE *l_free_rc_p;

  /* We have to free m_plugin_config_p and then the rendering
  ** context itself. */
  (*((FREE *)
     (l_rendering_context_p->m_plugin_p->m_funcptrs_p->m_delete_config_func)))
    (l_rendering_context_p->m_plugin_config_p);

  l_free_rc_p = l_rendering_context_p->m_free_func;
  (*l_free_rc_p)(l_rendering_context_p);
}

/*****************************************************************************
** Notify the plugin about a modification in the original text.
** The client-accessible function encapsulates a call to the plugin-provided
** function called p_notify_modification() and with the same arguments.
**
** The basic idea is that a change consists of deletion of a text segment
** starting at offset N and having length L1, followed by insertion
** of another text segment starting at the same offset and having length L2.
** Either text segment may have zero length.
** Thus, the same mechanism is used to describe both text replacements,
** insertions and deletions.
*****************************************************************************/

E_RENDER_RETURN_CODES
c_notify_modification(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
		      size_t in_change_is_at,
		                  /* The offset in the original text
                                  ** in which the change took place. */
		      size_t in_deleted_chars,
		                  /* Number of deleted characters */
		      size_t in_inserted_chars,
		                  /* Number of inserted characters */

		      GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
		      void *in_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the callback. */

		      size_t *out_first_modified_line_p)
		                  /* Reply to the client software - which is
				  ** the first rendered line to be affected
				  ** by the change. */
{
  P_NOTIFY_MODIFICATION *l_notify_modification_func;

  assert(NULL != in_out_text_p);
  assert(NULL != in_get_frag_func);
  assert(NULL != out_first_modified_line_p);

  assert(NULL != in_out_text_p->m_rendering_context_p);
  assert(NULL != in_out_text_p->m_rendering_context_p->m_plugin_p);
  assert(NULL != in_out_text_p->m_rendering_context_p->m_plugin_p
	 ->m_funcptrs_p);

  l_notify_modification_func = (P_NOTIFY_MODIFICATION *)
    (in_out_text_p->m_rendering_context_p->m_plugin_p
    ->m_funcptrs_p->m_notify_modification_func);
  assert(NULL != l_notify_modification_func);

  return((*l_notify_modification_func)(in_out_text_p,
				       in_change_is_at,
				       in_deleted_chars,
				       in_inserted_chars,
				       in_get_frag_func,
				       in_client_data_p,
				       out_first_modified_line_p));
}

/*****************************************************************************
** Retrieve a line of rendered glyphs.
** The client-accessible function encapsulates a call to the plugin-provided
** function called p_retrieve_rendered_line() and with the same arguments.
**
** Most plugin writers will probably implement the entire rendering
** algorithm in p_retrieve_rendered_line(), because this is the right place
** for lazy evaluation of the rendering algorithm.
*****************************************************************************/

E_RENDER_RETURN_CODES
c_retrieve_rendered_line(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
			 size_t in_desired_visual_row,
			          /* Index of the line which the client
				  ** software wants to retrieve. */
			 GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
			 void *in_frag_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_frag_func
				  ** callback. */
			 GET_RENDERED_LINE_SIZE *in_get_line_size_func,
			          /* Client-provided callback for
				  ** retrieving desired rendered line
				  ** size. */
			 void *in_size_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_line_size_func
				  ** callback. */
			 /* Output arguments - returned only if
			 ** out_glyphs_arr_p is not NULL. */
			 size_t *out_line_length_p,
			         /* The array length is in glyphs
				 ** (which could be multi-byte). */
			 void **out_glyphs_arr_p)
{
  P_RETRIEVE_RENDERED_LINE *l_retrieve_rendered_line_func;
  assert(NULL != in_out_text_p);
  assert(NULL != in_get_frag_func);
  assert(NULL != in_get_line_size_func);
  assert((NULL == out_glyphs_arr_p) || (NULL != out_line_length_p));
  assert((NULL == out_glyphs_arr_p) || (NULL == *out_glyphs_arr_p));

  assert(NULL != in_out_text_p->m_rendering_context_p);
  assert(NULL != in_out_text_p->m_rendering_context_p->m_plugin_p);
  assert(NULL != in_out_text_p->m_rendering_context_p->m_plugin_p
	 ->m_funcptrs_p);

  l_retrieve_rendered_line_func = (P_RETRIEVE_RENDERED_LINE *)
    (in_out_text_p->m_rendering_context_p->m_plugin_p
    ->m_funcptrs_p->m_retrieve_rendered_line_func);
  assert(NULL != l_retrieve_rendered_line_func);

  return ((*l_retrieve_rendered_line_func)(in_out_text_p,
					   in_desired_visual_row,
					   in_get_frag_func,
					   in_frag_client_data_p,
					   in_get_line_size_func,
					   in_size_client_data_p,
					   /* Output arguments */
					   out_line_length_p,
					   out_glyphs_arr_p));
}


/*****************************************************************************
** Given rendered text data structure (including rendering context pointer
** and other information), and cursor position in the original text
** ("logical" cursor position), find the visual cursor position (in terms
** of line and character position in line) and retrieve also the attributes
** of the glyph under the cursor.
**
** NOTE:  the procedure assumes that the text, for which cursor position
** translation is requested, has already been rendered.
** The return code RENDER_FAILURE_NOT_RENDERED_YET is returned if this
** assumption has been violated.  The client software may request to
** retrieve some more lines using c_retrieve_rendered_line() and then
** retry the operation.
*****************************************************************************/

E_RENDER_RETURN_CODES
l2v_cursor_translation(const SRenderedText *in_text_p,
		                  /* The rendered text data structure. */
		       size_t in_offset,
		                  /* Offset into the original string
				  ** of the character under the cursor. */
		       /* Output arguments */
		       size_t *out_visual_row_p,
		                  /* Number of the line in the rendered text,
				  ** on which the cursor resides.
				  ** The number of 1st line in the text is
				  ** zero. */
		       size_t *out_visual_col_p,
		                  /* Cursor X coordinate, normalized
				  ** into character position in row (starting
				  ** from zero). */
		       long *out_orig_frag_attributes_p,
		                  /* Attributes of the glyph under the
				  ** cursor, which were inherited from
				  ** the original text - such as BOLD
				  ** or UNDERLINE */
		       long *out_added_frag_attributes_p)
                                  /* Attributes added by the
				  ** plugin - such as LTR/RTL. */
{
  const SRenderedLine *l_current_line_p = NULL;
  const SRenderedFragment *l_current_fragment_p = NULL;
  E_RENDER_RETURN_CODES ret;

  assert(NULL != in_text_p);
  assert(NULL != out_visual_row_p);
  assert(NULL != out_visual_col_p);
  assert(NULL != out_orig_frag_attributes_p);
  assert(NULL != out_added_frag_attributes_p);

  ret = locate_fragment(in_text_p, in_offset,
			NULL, NULL,
			&l_current_line_p, &l_current_fragment_p);
  if (RENDER_SUCCESS != ret) return(ret);

  *out_visual_row_p =
    in_offset - l_current_fragment_p->m_first_orig_char_index;
  assert(*out_visual_row_p < l_current_fragment_p->m_orig_frag_length);

  assert(l_current_fragment_p->m_first_orig_char_index <= in_offset);
  assert(in_offset - l_current_fragment_p->m_first_orig_char_index
	 < l_current_fragment_p->m_orig_frag_length);
  *out_visual_col_p = l_current_fragment_p
    ->m_forward_xrefs_arr[in_offset
			 - l_current_fragment_p->m_first_orig_char_index];
  *out_orig_frag_attributes_p =
    l_current_fragment_p->m_orig_frag_attributes;
  *out_added_frag_attributes_p =
    l_current_fragment_p->m_added_frag_attributes;

  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Given rendered text data structure (including rendering context pointer
** and other information), and cursor position in the rendered text
** ("visual" cursor position), find the cursor position in the original
** text ("logical" cursor position).
** Also, retrieve also the attributes of the glyph under the visual cursor.
**
** NOTE:  the procedure assumes that the text, for which cursor position
** translation is requested, has already been rendered.
** The return code RENDER_FAILURE_NOT_RENDERED_YET is returned if this
** assumption has been violated.  The client software may request to
** retrieve some more lines using c_retrieve_rendered_line() and then
** retry the operation.
*****************************************************************************/

E_RENDER_RETURN_CODES
v2l_cursor_translation(const SRenderedText *in_text_p,
		                  /* The rendered text data structure. */
		       size_t in_visual_row,
		                  /* Number of the line in the rendered text,
				  ** on which the cursor resides.
				  ** The number of 1st line in the text is
				  ** zero. */
		       size_t in_visual_col,
		                  /* Cursor X coordinate, normalized
				  ** into character position in row (starting
				  ** from zero). */
		       /* Output arguments */
		       size_t *out_offset_p,
		                  /* Offset into the original string
				  ** of the character under the cursor. */
		       long *out_orig_frag_attributes_p,
		                  /* Attributes of the glyph under the
				  ** cursor, which were inherited from
				  ** the original text - such as BOLD
				  ** or UNDERLINE */
		       long *out_added_frag_attributes_p)
                                  /* Attributes added by the
				  ** plugin - such as LTR/RTL. */
{
  const SRenderedLine *l_current_line_p;
  const SRenderedFragment *l_current_fragment_p;
  size_t l_index;
  size_t l_col_offset;
  E_RENDER_RETURN_CODES ret;

  assert(NULL != in_text_p);
  assert(NULL != out_offset_p);
  assert(NULL != out_orig_frag_attributes_p);
  assert(NULL != out_added_frag_attributes_p);

  if (in_visual_row >= in_text_p->m_used_line_array_size) {
    /* The requested visual row does not exist in the array.
    ** Normally this means that the line was not rendered yet. */
    return(RENDER_FAILURE_NOT_RENDERED_YET);
  }
  l_current_line_p = in_text_p->m_line_p_arr[in_visual_row];

  /* Assert that there is at least one fragment in the line. */
  assert(l_current_line_p->m_used_frag_array_size > 0);

  /* Accumulate lengths of rendered text fragments until we
  ** reach in_visual_col. */
  for (l_index = 0, l_col_offset = 0;
       l_index < l_current_line_p->m_used_frag_array_size;
       l_index++) {
    l_current_fragment_p = l_current_line_p->m_frag_p_arr[l_index];
    if ((l_col_offset + l_current_fragment_p->m_frag_length)
	> in_visual_col) {
      /* Found the fragment in which the cursor resides. */
      size_t l_frag_offset;       /* Offset of the cursor relative
				  ** to fragment's first glyph. */
      assert(in_visual_col >= l_col_offset);
      l_frag_offset = in_visual_col - l_col_offset;
      assert(l_frag_offset < l_current_fragment_p->m_frag_length);

      *out_offset_p = l_current_fragment_p->m_back_xrefs_arr[l_frag_offset];
      *out_orig_frag_attributes_p =
	l_current_fragment_p->m_orig_frag_attributes;
      *out_added_frag_attributes_p =
	l_current_fragment_p->m_added_frag_attributes;

      return(RENDER_SUCCESS);
    }
    l_col_offset += l_current_fragment_p->m_frag_length;
  }
  /* If we are here, then the cursor is beyond any visual text on the
  ** line.  If this is not the last line of the document, then we'll
  ** want to return the data of the last glyph in the current line.
  ** If this IS the last line of the document, then we'll assume
  ** that the relevant text was not rendered yet. */
  if (in_visual_row == in_text_p->m_used_line_array_size - 1) {
    return(RENDER_FAILURE_NOT_RENDERED_YET);
  }
  assert(in_visual_row < in_text_p->m_used_line_array_size - 1);

  l_current_fragment_p = l_current_line_p->
    m_frag_p_arr[l_current_line_p->m_used_frag_array_size-1];

  *out_offset_p = l_current_fragment_p->
    m_back_xrefs_arr[l_current_fragment_p->m_frag_length-1];
  *out_orig_frag_attributes_p =
    l_current_fragment_p->m_orig_frag_attributes;
  *out_added_frag_attributes_p =
    l_current_fragment_p->m_added_frag_attributes;

  return(RENDER_SUCCESS);
}

/****************************************************************************/

/* End of plugclient0.c */
