/*****************************************************************************
**
** plugclient0.h
** $Header: /home/omer/work/cvs_repository/Hebrew/include/plugclient0.h,v 1.3 1999/02/02 01:23:33 omer Exp $
**
** Level 0 I18N text rendering plugin interface specifications, from the
** point of view of client applications.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

#ifndef _PLUGCLIENT0_H_
#define _PLUGCLIENT0_H_

#include <stdlib.h>
#include "structs0.h"
#include "plugimp0.h"

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************************
** bind_to_plugin()
** Finds the requested plugin and binds the shared library containing the
** plugin.
** The plugin will be configured by set_rendering_context().
*****************************************************************************/

E_RENDER_RETURN_CODES
bind_to_plugin(MALLOC *in_malloc_func,
	       FREE *in_free_func,
	                           /* Functions to be used for
				   ** allocating and freeing memory
				   ** as needed by this procedure.
				   ** No rendering context exists at
				   ** this point, so it can't be used.
				   ** Default values of malloc() and
				   ** free() are used for NULL arguments. */
	       const char *in_plugin_name_str,
	                           /* Name of plugin (typically, the name
				   ** of a file in a standard directory). */
	       SPlugin *out_plugin_p);
                                   /* Output argument, which will
				   ** receive bound plugin data. */

/*****************************************************************************
** The following function is used by bind_to_plugin() to validate the
** plugin type and version being used.
*****************************************************************************/

BOOL
validate_function_pointers(const SRenderingFunctions *in_rf_p);


/*****************************************************************************
** unbind_from_plugin()
** Releases the plugin.  If no one else is using the same plugin, the
** corresponding shared library is released from memory.
*****************************************************************************/

E_RENDER_RETURN_CODES
unbind_from_plugin(SPlugin *in_out_plugin_p);
                                   /* As input, this structure identifies
				   ** the shared library handle to be freed.
				   ** When done, all fields in this structure
				   ** are set to NULL. */

/*****************************************************************************
** Create a new rendering context structure and set its fields.
** This includes, wherever necessary, processing of default values.
** Optionally, an existing rendering context may be supplied in order
** to copy from it values not explicitly set.
*****************************************************************************/

/* Bitmask values for in_fields_mask below */
#define RC_PLUGIN           (1<<0)
#define RC_PLUGIN_CONFIG    (1<<1)
#define RC_MALLOC           (1<<2)
#define RC_FREE             (1<<3)
#define RC_GLYPH_SIZE       (1<<4)
#define RC_FRAG_PER_LINE    (1<<5)
#define RC_ALL              0xFFFFFFFF

E_RENDER_RETURN_CODES
set_clone_rendering_context
(const SRenderingContext *in_source_rc_p,
                                   /* Original rendering context, from which
				   ** we are cloning (possibly with
				   ** modifications) the new rendering
				   ** context.
				   ** If NULL, a brand new rendering
				   ** context is to be created. */
 long int in_fields_mask,          /* OR of bitmasks defining which
				   ** fields (if any) are to be modified
				   ** from the arguments.
				   ** If RC_ALL, in_source_rc_p is
				   ** ignored. */
 SPlugin *in_plugin_p,
 const char *in_config_str,        /* Plugin configuration string
				   ** (typically, the name of a
				   ** configuration file). */
 MALLOC *in_malloc_func,
 FREE *in_free_func,
 size_t in_bytes_per_glyph,
 size_t in_initial_total_frag_array_size,
 SRenderingContext **out_context_p_p);

/*****************************************************************************
** Free memory allocated to the fields of a rendering context structure.
*****************************************************************************/
FREE destroy_rendering_context;


/*****************************************************************************
** Notify the plugin about a modification in the original text.
** The client-accessible function encapsulates a call to the plugin-provided
** function called p_notify_modification() and with the same arguments.
**
** The basic idea is that a change consists of deletion of a text segment
** starting at offset N and having length L1, followed by insertion
** of another text segment starting at the same offset and having length L2.
** Either text segment may have zero length.
** Thus, the same mechanism is used to describe both text replacements,
** insertions and deletions.
*****************************************************************************/

E_RENDER_RETURN_CODES
c_notify_modification(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
		      size_t in_change_is_at,
		                  /* The offset in the original text
                                  ** in which the change took place. */
		      size_t in_deleted_chars,
		                  /* Number of deleted characters */
		      size_t in_inserted_chars,
		                  /* Number of inserted characters */

		      GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
		      void *in_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the callback. */

		      size_t *out_first_modified_line_p);
		                  /* Reply to the client software - which is
				  ** the first rendered line to be affected
				  ** by the change. */

/*****************************************************************************
** Retrieve a line of rendered glyphs.
** The client-accessible function encapsulates a call to the plugin-provided
** function called p_retrieve_rendered_line() and with the same arguments.
**
** Most plugin writers will probably implement the entire rendering
** algorithm in p_retrieve_rendered_line(), because this is the right place
** for lazy evaluation of the rendering algorithm.
*****************************************************************************/

E_RENDER_RETURN_CODES
c_retrieve_rendered_line(SRenderedText *in_out_text_p,
                                  /* The rendered text data structure, which
				  ** is affected by the announced change. */
			 size_t in_desired_visual_row,
			          /* Index of the line which the client
				  ** software wants to retrieve. */
			 GET_ORIG_TEXT_FRAGMENT *in_get_frag_func,
			          /* Client-provided callback for
			          ** retrieving original text
			          ** fragments. */
			 void *in_frag_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_frag_func
				  ** callback. */
			 GET_RENDERED_LINE_SIZE *in_get_line_size_func,
			          /* Client-provided callback for
				  ** retrieving desired rendered line
				  ** size. */
			 void *in_size_client_data_p,
			          /* Client data, which is passed as
			          ** it is to the in_get_line_size_func
				  ** callback. */
			 /* Output arguments */
			 size_t *out_line_length_p,
			         /* The array length is in glyphs
				 ** (which could be multi-byte). */
			 void **out_glyphs_arr_p);

/*****************************************************************************
** Given rendered text data structure (including rendering context pointer
** and other information), and cursor position in the original text
** ("logical" cursor position), find the visual cursor position (in terms
** of line and character position in line) and retrieve also the attributes
** of the glyph under the cursor.
**
** NOTE:  the procedure assumes that the text, for which cursor position
** translation is requested, has already been rendered.
** The return code RENDER_FAILURE_NOT_RENDERED_YET is returned if this
** assumption has been violated.  The client software may request to
** retrieve some more lines using c_retrieve_rendered_line() and then
** retry the operation.
*****************************************************************************/

E_RENDER_RETURN_CODES
l2v_cursor_translation(const SRenderedText *in_text_p,
		                  /* The rendered text data structure. */
		       size_t in_offset,
		                  /* Offset into the original string
				  ** of the character under the cursor. */
		       /* Output arguments */
		       size_t *out_visual_row_p,
		                  /* Number of the line in the rendered text,
				  ** on which the cursor resides.
				  ** The number of 1st line in the text is
				  ** zero. */
		       size_t *out_visual_col_p,
		                  /* Cursor X coordinate, normalized
				  ** into character position in row (starting
				  ** from zero). */
		       long *out_orig_frag_attributes_p,
		                  /* Attributes of the glyph under the
				  ** cursor, which were inherited from
				  ** the original text - such as BOLD
				  ** or UNDERLINE */
		       long *out_added_frag_attributes_p);
                                  /* Attributes added by the
				  ** plugin - such as LTR/RTL. */

/*****************************************************************************
** Given rendered text data structure (including rendering context pointer
** and other information), and cursor position in the rendered text
** ("visual" cursor position), find the cursor position in the original
** text ("logical" cursor position).
** Also, retrieve also the attributes of the glyph under the visual cursor.
**
** NOTE:  the procedure assumes that the text, for which cursor position
** translation is requested, has already been rendered.
** The return code RENDER_FAILURE_NOT_RENDERED_YET is returned if this
** assumption has been violated.  The client software may request to
** retrieve some more lines using c_retrieve_rendered_line() and then
** retry the operation.
*****************************************************************************/

E_RENDER_RETURN_CODES
v2l_cursor_translation(const SRenderedText *in_text_p,
		                  /* The rendered text data structure. */
		       size_t in_visual_row,
		                  /* Number of the line in the rendered text,
				  ** on which the cursor resides.
				  ** The number of 1st line in the text is
				  ** zero. */
		       size_t in_visual_col,
		                  /* Cursor X coordinate, normalized
				  ** into character position in row (starting
				  ** from zero). */
		       /* Output arguments */
		       size_t *out_offset_p,
		                  /* Offset into the original string
				  ** of the character under the cursor. */
		       long *out_orig_frag_attributes_p,
		                  /* Attributes of the glyph under the
				  ** cursor, which were inherited from
				  ** the original text - such as BOLD
				  ** or UNDERLINE */
		       long *out_added_frag_attributes_p);
                                  /* Attributes added by the
				  ** plugin - such as LTR/RTL. */

/****************************************************************************/

#ifdef __cplusplus
}; /* end of extern "C" */
#endif

#endif /* _PLUGCLIENT0_H_ */

/* End of plugclient0.h */
