/*****************************************************************************
**
** plugfilter0.c
*/
static char st_plugfilter0_version[] = "$Header: /home/omer/work/cvs_repository/Hebrew/include/plugfilter0.c,v 1.6 1999/02/06 18:22:11 omer Exp $";
/*
** A simple filter, which reads data from stdin (until EOF or 1MB
** have been read).  Then it renders it using an user-selected level
** 0 text rendering plugin.  The rendered text is then written to
** stdout.
**
** Copyright (C) 1999  Omer Zak
**
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** with this program; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/plugclient0.h"

/*****************************************************************************
** Auxiliary procedures                                                     **
*****************************************************************************/
void
usage(int argc, char *argv[])
{
  /* Print usage message and exit (does not return to the caller) */
  fprintf(stderr, "Usage: %s plugin_name [plugin_option_string] < input_file > output_file\n", argv[0]);
  exit(1);
}

/*****************************************************************************
** Callback for retrieving original text fragments                          **
*****************************************************************************/

ssize_t st_orig_text_length = 0;  /* This variable was made global because
				  ** I wanted to avoid the hassle of passing
				  ** more complicated data structure in
				  ** in_client_data_p, reserving it instead
				  ** only for the big text buffer itself. */

E_RENDER_RETURN_CODES
get_frag_func
(const void *in_client_data_p,
                                  /* Client-provided data, used by the
				  ** client callback to identify the context
				  ** in which the callback is to operate
				  ** (such as from which text string
				  ** to retrieve the data). */
 size_t in_origtext_offset,
			          /* Offset inside the original
				  ** string where we want the fragment
				  ** to start. */
 /* The following are output parameters */
 size_t *out_frag_length_p,
			          /* Length of the returned fragment
				  ** (guaranteed only to be > 0). */
 long *out_orig_frag_attributes_p,
			          /* Original fragment attributes, to be
				  ** saved in the SRenderedFragment field
				  ** m_orig_frag_attributes. */
 char **out_frag_str_p)
                                  /* Pointer at a pointer which will be set
                                  ** to point at the returned fragment. */
{
  assert(NULL != out_frag_length_p);
  assert(NULL != out_orig_frag_attributes_p);
  assert(NULL != out_frag_str_p);
  assert(NULL == *out_frag_str_p);

  if (in_origtext_offset >= (size_t) st_orig_text_length) {
    /* We attempted to retrieve a fragment beyond the end of the
    ** original text.
    ** The caller should now perform any relevant end-of-text processing. */
    *out_frag_length_p = 0;
    *out_orig_frag_attributes_p = 0;
    *out_frag_str_p = NULL;
    return(RENDER_BEYOND_END_OF_TEXT);
  }

  *out_frag_length_p = (size_t) st_orig_text_length - in_origtext_offset;
  *out_orig_frag_attributes_p = 0; /* No original attributes at present */
  *out_frag_str_p = ((char *)in_client_data_p) + in_origtext_offset;

  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Callback for retrieving rendered line length [in glyphs]                 **
*****************************************************************************/
E_RENDER_RETURN_CODES
line_size_func
(const void *in_client_data_p,
                                  /* Client-provided data, used by the
				  ** client callback to identify the context
				  ** in which the callback is to operate
				  ** (such as from which text string
				  ** to retrieve the data). */
 size_t in_visual_row,
			          /* Number of the line in the rendered
				  ** text for which we want to know the
				  ** length desired by the client software.
				  ** The number of 1st line in the text is
				  ** zero. */
 /* The following are output parameters */
 size_t *out_line_length_p)
                                  /* Desired line length. */
{
  assert(NULL != out_line_length_p);
  *out_line_length_p = (size_t) in_client_data_p;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Main procedure                                                           **
*****************************************************************************/
int
main(int argc, char *argv[])
{
  /* Local variables */
  E_RENDER_RETURN_CODES ret;
  SPlugin l_plugin = {NULL,NULL};
  SRenderingContext *l_rc_p = NULL;

  char *l_big_buf = NULL;    /* Big buffer for the original text */

  SRenderedText *l_rendered_text_p = NULL;
  size_t l_first_modified_line;
  size_t l_visual_row;
  size_t l_line_length;
  void *l_line_text_p;
  ssize_t l_actually_written_text_length;

  /* !!!!! Register here an at_exit() to free all dynamic memory !!!!*/

  /* Check command line options */
  if (argc < 2) {
    fprintf(stderr, "Missing plugin name.\n");
    usage(argc,argv);
  }
  if (argc > 3) {
    fprintf(stderr, "Only one configuration string is allowed.\n");
    usage(argc,argv);
  }

  ret = bind_to_plugin(NULL, NULL,
		       argv[1],     /* in_plugin_name_str */
		       &l_plugin);
  if (RENDER_SUCCESS != ret) {
    fprintf(stderr,"Attempt to bind to plugin %s failed due to reason %d.\n",
	    argv[1], (int) ret);
    exit(1);
  }

  ret = set_clone_rendering_context(NULL, RC_ALL,
				    &l_plugin,
				    ((argc == 2) ? NULL : argv[2]),
				    NULL,    /* default malloc() */
				    NULL,    /* default free() */
				    1,       /* bytes per glyph */
				    3,       /* start with 3 fragments/line */
				    &l_rc_p);
  if (RENDER_SUCCESS != ret) {
    fprintf(stderr,"Attempt to create rendering context with configuration %s"
	    " failed due to reason %d.\n",
	    ((argc == 2) ? "<NULL>" : argv[2]),
	    (int) ret);
  }
  assert(NULL != l_rc_p);
  /* End of plugin clients' prologue code */

  /* This specific plugin client merely reads data from stdin,
  ** passes it via an user-selectable plugin, and writes the output
  ** to stdout. */

  /* Allocate memory and read the entire stdin into memory.
  ** We can handle files whose length is not greater than 1MB
  ** (an arbitrary limitation). */

#define CRAZY_FILE_SIZE 1048576

  l_big_buf = (char *) malloc(CRAZY_FILE_SIZE+10);
  if (l_big_buf == NULL) {
    fprintf(stderr,"Could not allocate memory to read data from stdin.\n");
    exit(1);
  }

  st_orig_text_length = read(0 /* stdin */, l_big_buf, CRAZY_FILE_SIZE);
  if (st_orig_text_length < 0) {
    perror("Error while trying to read from stdin");
    exit(1);
  }
  else if (st_orig_text_length == 0) {
    fprintf(stderr,"Empty stdin.\n");
    exit(1);
  }

  l_rendered_text_p = construct_rendered_text(l_rc_p, 25);
  if (NULL == l_rendered_text_p) {
    fprintf(stderr,"Failed to allocate SRenderedText structure.\n");
    exit(1);
  }

  /* Now we have stuff to render. */
  ret = c_notify_modification(l_rendered_text_p,
			      0,
			      0,
			      (size_t) st_orig_text_length,
			      &get_frag_func,
			      l_big_buf,     /* Client data */
			      &l_first_modified_line);
  if (RENDER_SUCCESS != ret) {
    fprintf(stderr,"c_notify_modification(orig text length = %u) failed"
	    " due to reason %d.\n",
	    (unsigned int) st_orig_text_length,
	    (int) ret);
    exit(1);
  }
  assert(0 == l_first_modified_line);
  /* We notified a modification at offset 0, so we expect that the
  ** first line to be reported as invalidated by the plugin is line 0. */

  /* After having notified the plugin about modifications to the
  ** original text, we may start retrieving the rendered text.
  ** The plugin performs the actual rendering some time between
  ** the modification notification and the retrieval operation. */

  for (l_visual_row = 0;;l_visual_row++) {
    l_line_text_p = NULL;

    ret = c_retrieve_rendered_line(l_rendered_text_p,
				   l_visual_row,
				   &get_frag_func,
				   l_big_buf,  /* Client data */
				   &line_size_func,
				   (void *) 74,  /* Client data for
						 ** line_size_func() */
				   &l_line_length,
				   &l_line_text_p);
    if (RENDER_BEYOND_END_OF_TEXT == ret) {
      break;
    }
    else if (RENDER_SUCCESS != ret) {
      fprintf(stderr,"Attempt to retrieve visual line %d failed"
	      " due to reason %d.\n",
	      (int) l_visual_row, (int) ret);
      exit(1);
    }

    assert(l_line_length <= 74);

    l_actually_written_text_length =
      write(1 /* stdout */, l_line_text_p, l_line_length);
    if (l_actually_written_text_length < 0) {
      perror("Error while trying to write to stdout");
      exit(1);
    }
    assert(l_line_length == (size_t) l_actually_written_text_length);
    free(l_line_text_p);
  }

  /* Finished doing the actual work.  Now free all allocated resources. */

  destroy_rendered_text(l_rendered_text_p);
  free(l_big_buf);

  /* The following are needed in all plugin clients. */
  destroy_rendering_context(l_rc_p);
  ret = unbind_from_plugin(&l_plugin);
  if (RENDER_SUCCESS != ret) {
    fprintf(stderr,
	    "Attempt to unbind from plugin %s failed due to reason %d.\n",
	    argv[1], (int) ret);
    exit(1);
  }

  return(0);  /* Normal termination */
}

/****************************************************************************/

/* End of plugfilter0.c */
