/*****************************************************************************
**
** plugimp0.c
*/
static char st_plugimp0_version[] = "$Header: /home/omer/work/cvs_repository/Hebrew/include/plugimp0.c,v 1.3 1999/02/06 18:22:11 omer Exp $";
/*
** Level 0 I18N text rendering plugin auxiliary procedures.
** Those procedures are invoked by both client software and
** plugins.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

/* #define DEBUG */

#include <assert.h>
#include "structs0.h"
#include "plugimp0.h"

/*****************************************************************************
** Validate rendering context - used by both plugin and client software.
**
** Return values:  FALSE - something is wrong.
**                 TRUE - all is okay.
*****************************************************************************/
BOOL
validate_rendering_context(const SRenderingContext *in_context_p)
{
  if (NULL == in_context_p->m_plugin_p) return(FALSE);
  if (NULL == in_context_p->m_plugin_p->m_funcptrs_p) return(FALSE);

  /* in_context_p->m_plugin_config_p - will be validated by
  ** the plugin itself */

  if (NULL == in_context_p->m_malloc_func) return(FALSE);

  if (NULL == in_context_p->m_free_func) return(FALSE);

  if (0 == in_context_p->m_bytes_per_glyph) return(FALSE);
  if (in_context_p->m_bytes_per_glyph > 16) return(FALSE);
  /* The maximum of 16 is a mere sanity check.  For Mongolian it
  ** might be higher if the glyphs will be bit-mapped. */

  if (0 == in_context_p->m_initial_total_frag_array_size) return(FALSE);

  return(TRUE);
}

/*****************************************************************************
** Fragment destructor
*****************************************************************************/
void
destroy_fragment(void *in_p)
{
  SRenderedFragment *l_frag_p = (SRenderedFragment *) in_p;
  FREE *l_free_func;

  assert(NULL != l_frag_p);
  l_free_func = l_frag_p->m_frag_free_func;
  assert(NULL != l_free_func);

  assert(NULL != l_frag_p->m_frag_str);
  l_free_func(l_frag_p->m_frag_str);

  assert(NULL != l_frag_p->m_back_xrefs_arr);
  l_free_func(l_frag_p->m_back_xrefs_arr);

  if (l_frag_p->m_orig_frag_length > 0) {
    assert(NULL != l_frag_p->m_forward_xrefs_arr);
    l_free_func(l_frag_p->m_forward_xrefs_arr);
  } else {
    assert(NULL == l_frag_p->m_forward_xrefs_arr);
  }

  if (NULL != l_frag_p->m_frag_state_free_func) {
    l_frag_p->m_frag_state_free_func(l_frag_p->m_frag_state_p);
  }

  /* Now we can free the structure itself. */
  l_free_func(l_frag_p);
}

/*****************************************************************************
** Line destructor
*****************************************************************************/
void
destroy_line(void *in_p)
{
  SRenderedLine *l_line_p = (SRenderedLine *) in_p;
  FREE *l_free_func;
  size_t l_index;

  assert(NULL != l_line_p);
  l_free_func = l_line_p->m_line_free_func;
  assert(NULL != l_free_func);

  /* Destroy all fragments */
  assert(l_line_p->m_total_frag_array_size > 0);
  assert(l_line_p->m_total_frag_array_size
	 >= l_line_p->m_used_frag_array_size);
  assert(NULL != l_line_p->m_frag_p_arr);

  for (l_index = 0; l_index < l_line_p->m_used_frag_array_size; l_index++) {
    assert(NULL != l_line_p->m_frag_p_arr[l_index]);
    destroy_fragment((void *)l_line_p->m_frag_p_arr[l_index]);
  }

  l_free_func(l_line_p->m_frag_p_arr);

  /* Now we can free the structure itself. */
  l_free_func(l_line_p);
}

/*****************************************************************************
** Rendered text data structure - constructor and destructor.
** The constructor returns NULL if memory allocation attempt failed.
*****************************************************************************/
SRenderedText *
construct_rendered_text(const SRenderingContext *in_rc_p,
			size_t in_initial_total_line_array_size)
{
  SRenderedText *l_new_text_p = NULL;
  size_t l_index;          /* Loop index */

  assert(NULL != in_rc_p);
  assert(validate_rendering_context(in_rc_p));
  assert(in_initial_total_line_array_size > 0);

  l_new_text_p =
    (SRenderedText *)(in_rc_p->m_malloc_func(sizeof(SRenderedText)));
  if (NULL == l_new_text_p) return((SRenderedText *) NULL);

  /* Set fields in the new structure. */
  l_new_text_p->m_text_free_func = in_rc_p->m_free_func;
  l_new_text_p->m_total_line_array_size = in_initial_total_line_array_size;
  l_new_text_p->m_used_line_array_size = 0;
  l_new_text_p->m_line_p_arr = in_rc_p->
    m_malloc_func(sizeof(SRenderedLine *) * in_initial_total_line_array_size);
  l_new_text_p->m_rendering_context_p = in_rc_p;

  /* Check if any attempt to allocate memory failed. */
  if (NULL == l_new_text_p->m_line_p_arr) {
    /* We ran out of memory, so clean up and return NULL. */
    in_rc_p->m_free_func(l_new_text_p);
    return((SRenderedText *) NULL);
  }

  /* Initialize l_new_text_p->m_line_p_arr. */
  for (l_index = 0; l_index < in_initial_total_line_array_size; l_index++) {
    l_new_text_p->m_line_p_arr[l_index] = NULL;
  }

  /* All is now OK.  The structure is fully initialized.  The caller
  ** needs only to add at least one line to the structure. */
  return(l_new_text_p);
}

/*****************************************************************************
** Locate the fragment, which contains the glyph corresponding to
** a given logical offset.
*****************************************************************************/
E_RENDER_RETURN_CODES
locate_fragment(const SRenderedText *in_text_p,
		size_t in_offset, /* Offset into the original string
				  ** of the character under the cursor. */
		/* Output arguments */
		size_t *out_current_line_index_p,
		size_t *out_current_frag_index_p,
		const SRenderedLine **out_current_line_p_p,
		const SRenderedFragment **out_current_fragment_p_p)
{
  size_t l_line_index;
  size_t l_frag_index;
  SRenderedLine *l_curr_line_p;
  SRenderedFragment *l_curr_frag_p;

  assert(NULL != in_text_p);
  /* We are not validating the rest of *in_text_p. */
  assert((NULL == out_current_line_p_p) || (NULL == *out_current_line_p_p));
  assert((NULL == out_current_fragment_p_p)
	 || (NULL == *out_current_fragment_p_p));

  /* The present implementation uses linear search.
  ** !!! Should be re-implemented using binary search. */

  for (l_line_index = 0;
       l_line_index < in_text_p->m_used_line_array_size;
       l_line_index++) {
    l_curr_line_p = in_text_p->m_line_p_arr[l_line_index];
    /* Retrieve the last fragment of the current line. */
    assert(l_curr_line_p->m_used_frag_array_size > 0);
    assert(l_curr_line_p->m_frag_p_arr[0]->m_first_orig_char_index
	   <= in_offset);

    l_curr_frag_p =
      l_curr_line_p->m_frag_p_arr[l_curr_line_p->m_used_frag_array_size - 1];

    if (l_curr_frag_p->m_first_orig_char_index
	+ l_curr_frag_p->m_orig_frag_length > in_offset) {
      /* We located the line.  Now locate the fragment inside the line. */
      for (l_frag_index = 0;
	   l_frag_index < l_curr_line_p->m_used_frag_array_size;
	   l_frag_index++) {
	l_curr_frag_p = l_curr_line_p->m_frag_p_arr[l_frag_index];
	assert(l_curr_frag_p->m_first_orig_char_index <= in_offset);
	if (l_curr_frag_p->m_first_orig_char_index
	    + l_curr_frag_p->m_orig_frag_length > in_offset) {
	  /* The desired fragment was located.  Now we can return
	  ** the results to the caller. */
	  if (NULL != out_current_line_index_p) {
	    *out_current_line_index_p = l_line_index;
	  }
	  if (NULL != out_current_frag_index_p) {
	    *out_current_frag_index_p = l_frag_index;
	  }
	  if (NULL != out_current_line_p_p) {
	    *out_current_line_p_p = l_curr_line_p;
	  }
	  if (NULL != out_current_fragment_p_p) {
	    *out_current_fragment_p_p = l_curr_frag_p;
	  }
	  return(RENDER_SUCCESS);
	}
      }
      assert(FALSE);     /* We located the line but not the fragment.
			 ** This means that something is very wrong. */
    }
  }

  return(RENDER_FAILURE_NOT_RENDERED_YET);
}

/*****************************************************************************
** Destroy rendered text data structure.
*****************************************************************************/

void
destroy_rendered_text(void *in_p)
{
  SRenderedText *l_text_p = (SRenderedText *) in_p;
  FREE *l_free_func;
  size_t l_index;

  assert(NULL != l_text_p);
  /* Validate the SRenderedText structure contents. */
  assert(NULL != l_text_p->m_text_free_func);
  assert(l_text_p->m_total_line_array_size > 0);
  assert(l_text_p->m_total_line_array_size
	 >= l_text_p->m_used_line_array_size);
  assert(NULL != l_text_p->m_line_p_arr);
  assert(NULL != l_text_p->m_rendering_context_p);
  assert(validate_rendering_context(l_text_p->m_rendering_context_p));

  l_free_func = l_text_p->m_text_free_func;
  assert(l_free_func == l_text_p->m_rendering_context_p->m_free_func);

  /* Destroy all lines */
  for (l_index = 0; l_index < l_text_p->m_used_line_array_size; l_index++) {
    assert(NULL != l_text_p->m_line_p_arr[l_index]);
    destroy_line((void *)l_text_p->m_line_p_arr[l_index]);
  }

  l_free_func(l_text_p->m_line_p_arr);

  /* Now we can free the structure itself. */
  l_free_func(l_text_p);
}

/*****************************************************************************
** Dump (part of) rendered text data structure.
*****************************************************************************/
#ifdef DEBUG
void
dump_SRenderedText(const SRenderedText *in_rt_p,
		   FILE *out_file)
{
  const SRenderedLine *l_line_p;
  const SRenderedFragment *l_frag_p;
  size_t l_line_index;
  size_t l_frag_index;

  assert(NULL != in_rt_p);
  assert(NULL != out_file);

  fprintf(out_file,"Dumping SRenderedText structure\n"
	           "-------------------------------\n"
	  "total line array size: %u\n"
	  "used line array size:  %u\n",
	  (unsigned int) in_rt_p->m_total_line_array_size,
	  (unsigned int) in_rt_p->m_used_line_array_size);
  fprintf(out_file,"Rendering context:  bytes/glyph=%u,"
	  " initial frag array size=%u\n",
	  (unsigned int) in_rt_p->m_rendering_context_p->m_bytes_per_glyph,
	  (unsigned int) in_rt_p->m_rendering_context_p
	                                ->m_initial_total_frag_array_size);

  for (l_line_index = 0; l_line_index < in_rt_p->m_used_line_array_size;
       l_line_index++) {
    l_line_p = in_rt_p->m_line_p_arr[l_line_index];
    assert(NULL != l_line_p);

    fprintf(out_file,"\nLine %u:  row width=%u, total frag array=%u,"
	    " used fragments=%u\n",
	    (unsigned int) l_line_index,
	    (unsigned int) l_line_p->m_screen_row_width,
	    (unsigned int) l_line_p->m_total_frag_array_size,
	    (unsigned int) l_line_p->m_used_frag_array_size);
    for (l_frag_index = 0; l_frag_index < l_line_p->m_used_frag_array_size;
	 l_frag_index++) {
      l_frag_p = l_line_p->m_frag_p_arr[l_frag_index];
      assert(NULL != l_frag_p);

      fprintf(out_file,"    frag %u:  rendered length=%u\n"
                       "              attributes=%ld  added attributes=%ld\n",
	      (unsigned int) l_frag_index,
	      (unsigned int) l_frag_p->m_frag_length,
	      l_frag_p->m_orig_frag_attributes,
	      l_frag_p->m_added_frag_attributes);
      fprintf(out_file,"              1st orig char index=%u, "
	      "orig frag length=%u\n",
	      (unsigned int) l_frag_p->m_first_orig_char_index,
	      (unsigned int) l_frag_p->m_orig_frag_length);
    }
  }
}

#endif /* DEBUG */

/****************************************************************************/

/* End of plugimp0.c */
