/*****************************************************************************
**
** plugimp0.h
** $Header: /home/omer/work/cvs_repository/Hebrew/include/plugimp0.h,v 1.7 1999/02/06 18:22:11 omer Exp $
**
** Level 0 I18N text rendering plugin auxiliary procedures.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
******************************************************************************
** Responsibilities for implementing the various functions of the
** level 0 plugin interface:
** --------------------------------------------------------------
** struct0.h  - declares data structures
** plugimp0.h - declares library functions, to be used by both plugin
**              writers and client software developers.
** plugin0.h  - declares functions to be implemented by the plugin writers.
** plugclient0.h - declares the interface exposed to client software i.e.
**              functions to be implemented by the client software
**              developers and functions, which the plugin exports to
**              the client software.
**
** In the following:
** library    - function to be used by both plugin and client software.
** library/c  - function to be used only by client software.
** library/p  - function to be used only by plugin.
**
** Function                        Supplied by:   Used by:
** ------------------------------- -------------- ----------------------
** _init()                         Plugin writer  dlopen()
** _fini()                         Plugin writer  dlclose()
**
** bind_to_plugin()                library/c      Client software
** return_function_pointers()      Plugin writer  bind_to_plugin()
** validate_function_pointers()    library/c      bind_to_plugin()
** unbind_from_plugin()            library/c      Client software
**
** parse_config()                  Plugin writer  set_clone_rendering_context()
** validate_config()               Plugin writer  Plugin (internally)
** delete_config()                 Plugin writer  destroy_rendering_context()
**
** set_clone_rendering_context()   library/c      Client software
** validate_rendering_context()    library        library
** destroy_rendering_context()     library/c      Client software
**
** construct_fragment()            library/p      Plugin writer
** destroy_fragment()              library        Plugin writer,Client S/W
** construct_line()                library/p      Plugin writer
** line_append_fragment()          library/p      Plugin writer
** line_delete_fragments()         library/p      Plugin writer
** destroy_line()                  library        Plugin writer,Client S/W
** construct_rendered_text()       library        Plugin writer,Client S/W
** text_append_line()              library/p      Plugin writer
** text_delete_lines()             library/p      Plugin writer
** locate_origtext_offset()        library/p      Plugin writer
** locate_fragment()               library        Plugin writer,Client S/W
** destroy_rendered_text()         library        Plugin writer,Client S/W
** dump_SRenderedText()            library/DEBUG  Plugin writer,Client S/W
**
** notify_modification()           Plugin writer  Client software
** retrieve_rendered_line()        Plugin writer  Client software
** get_orig_text_fragment()        Client S/W     Plugin (callback)
** get_rendered_line_size()        Client S/W     Plugin (callback)
**
** l2v_cursor_translation()        library/c      Client software
** v2l_cursor_translation()        library/c      Client software
**
** Source code files to be provided
** --------------------------------
** plugimp0.c - implements library procedures (plugin + client)
** plugclient0.c - implements library/c procedures (client only)
** plugin0.c  - implements library/p procedures (plugin only)
**
** When linking plugins, you should link your own plugin code with
** plugimp0.o and plugin0.o.
** When linking client software, you should link your own software
** with plugimp0.o and plugclient0.o.
**
*****************************************************************************/

#ifndef _PLUGIMP0_H_
#define _PLUGIMP0_H_

#include <stdlib.h>
#include "structs0.h"

#ifdef DEBUG
#include <stdio.h>
#endif /* DEBUG */

#ifdef __cplusplus
extern "C" {
#endif


/*****************************************************************************
** The following are the main connection point between the plugin
** and the client software.
** Due to the existence of enveloping library procedures (see
** plugclient0.h, plugclient0.c), normal client software does not
** manipulate directly the structures and functions in question.
*****************************************************************************/

typedef const SRenderingFunctions *(RETURN_FUNCTION_POINTERS)(void);

/*****************************************************************************
** Validate rendering context - used by both plugin and client software.
**
** Return values:  FALSE - something is wrong.
**                 TRUE - all is okay.
*****************************************************************************/
BOOL
validate_rendering_context(const SRenderingContext *in_context_p);

/*****************************************************************************
** Fragment destructor
*****************************************************************************/
FREE destroy_fragment;
/*****************************************************************************
** Line destructor
*****************************************************************************/
FREE destroy_line;


/*****************************************************************************
** Rendered text data structure - constructor.
** The constructor returns NULL if memory allocation attempt failed.
*****************************************************************************/
SRenderedText *
construct_rendered_text(const SRenderingContext *in_rc_p,
			size_t in_initial_total_line_array_size);

/*****************************************************************************
** Locate the fragment, which contains the glyph corresponding to
** a given logical offset.
** Output arguments, in which the client is not interested, may
** be set to NULL.
*****************************************************************************/
E_RENDER_RETURN_CODES
locate_fragment(const SRenderedText *in_text_p,
		size_t in_offset, /* Offset into the original string
				  ** of the character under the cursor. */
		/* Output arguments */
		size_t *out_current_line_index_p,
		size_t *out_current_frag_index_p,
		const SRenderedLine **out_current_line_p_p,
		const SRenderedFragment **out_current_fragment_p_p);

/*****************************************************************************
** Destroy rendered text data structure.
*****************************************************************************/
FREE destroy_rendered_text;

/*****************************************************************************
** Dump (part of) rendered text data structure.
*****************************************************************************/
#ifdef DEBUG
void
dump_SRenderedText(const SRenderedText *in_rt_p,
		   FILE *out_file);
#endif /* DEBUG */

/*****************************************************************************
** GET_ORIG_TEXT_FRAGMENT data type is the type of a callback function,
** which the client must provide to the plugin.
** The plugin calls this function as needed to retrieve original text
** fragments.
*****************************************************************************/

typedef E_RENDER_RETURN_CODES
(GET_ORIG_TEXT_FRAGMENT)(const void *in_client_data_p,
                                  /* Client-provided data, used by the
				  ** client callback to identify the context
				  ** in which the callback is to operate
				  ** (such as from which text string
				  ** to retrieve the data). */
			 size_t in_origtext_offset,
			          /* Offset inside the original
				  ** string where we want the fragment
				  ** to start. */
			 /* The following are output parameters */
			 size_t *out_frag_length_p,
			          /* Length of the returned fragment
				  ** (guaranteed only to be > 0). */
			 long *out_orig_frag_attributes_p,
			          /* Original fragment attributes, to be
				  ** saved in the SRenderedFragment field
				  ** m_orig_frag_attributes. */
			 char **out_frag_str_p);
                                  /* Pointer at a pointer which will be set
                                  ** to point at the returned fragment. */

/* Notes:
** 1. This callback may be called by notify_modification() and/or
**    retrieve_rendered_line() at any time.
** 2. As long as a notify_modification() call or a retrieve_rendered_line()
**    call is in progress, the client software must not move fragments
**    whose details were returned to the plugin by means of
**    get_orig_text_fragment() calls.
**    However, the permanent data structures refer to original text
**    fragments by means of offsets, so client software is free to move
**    them around once the above plugin functions have exitted.
** 3. If the requested fragment lies beyond end of the original text,
**    then NULL pointer and zero-length fragment are returned.
**    In addition, the return code is set to RENDER_BEYOND_END_OF_TEXT. */

/*****************************************************************************
** GET_RENDERED_LINE_SIZE data type is the type of a callback function,
** which the client must provide to the plugin.
** The plugin calls this function as needed to retrieve the desired
** rendered line length (in glyphs), given a line serial number.
** Simple clients will just return a fixed value in the callback.
** The support exists to accommodate clients, which display the rendered
** text in windows having complicated geometry.
*****************************************************************************/

typedef E_RENDER_RETURN_CODES
(GET_RENDERED_LINE_SIZE)(const void *in_client_data_p,
                                  /* Client-provided data, used by the
				  ** client callback to identify the context
				  ** in which the callback is to operate
				  ** (such as from which text string
				  ** to retrieve the data). */
			 size_t in_visual_row,
			          /* Number of the line in the rendered
				  ** text for which we want to know the
				  ** length desired by the client software.
				  ** The number of 1st line in the text is
				  ** zero. */
			 /* The following are output parameters */
			 size_t *out_line_length_p);
                                  /* Desired line length. */

/****************************************************************************/

#ifdef __cplusplus
}; /* end of extern "C" */
#endif

#endif /* _PLUGIMP0_H_ */

/* End of plugimp0.h */
