/*****************************************************************************
**
** plugin0.c
*/
static char st_plugin0_version[] = "$Header: /home/omer/work/cvs_repository/Hebrew/include/plugin0.c,v 1.3 1999/02/06 16:49:26 omer Exp $";
/*
** Implements some procedures, which are needed by level 0 text
** rendering plugins.
** Those procedures are invoked by the plugins themselves.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

#include <assert.h>
#include "structs0.h"
#include "plugimp0.h"
#include "plugin0.h"
#include "plugclient0.h"


/*****************************************************************************
** Fragment constructor
** NULL is returned if memory allocation attempt failed.
**
** Not all SRenderedFragment fields and data structures are initialized
** by this constructor.
** The calling software is responsible for initializing some of them.
*****************************************************************************/

SRenderedFragment *
construct_fragment(const SRenderingContext *in_rc_p,
		   size_t in_frag_length,
		   long in_orig_frag_attributes,
		   long in_added_frag_attributes,
		   size_t in_first_orig_char_index,
		   size_t in_orig_frag_length)
{
  SRenderedFragment *l_new_frag_p = NULL;

  assert(NULL != in_rc_p);
  assert(validate_rendering_context(in_rc_p));
  assert(in_frag_length > 0);
  /* It is possible for in_orig_frag_length to be zero, if the
  ** plugin inserts, on its initiative, a fragment with different
  ** attributes than its surroundings. */

  l_new_frag_p =
    (SRenderedFragment *) (in_rc_p->m_malloc_func(sizeof(SRenderedFragment)));
  if (NULL == l_new_frag_p) return(NULL);

  /* Set fields in the new fragment structure. */
  l_new_frag_p->m_frag_free_func = in_rc_p->m_free_func;
  l_new_frag_p->m_frag_length = in_frag_length;
  l_new_frag_p->m_orig_frag_attributes = in_orig_frag_attributes;
  l_new_frag_p->m_added_frag_attributes = in_added_frag_attributes;
  l_new_frag_p->m_frag_str =
    in_rc_p->m_malloc_func(in_frag_length * in_rc_p->m_bytes_per_glyph);
  l_new_frag_p->m_back_xrefs_arr =
    in_rc_p->m_malloc_func(sizeof(size_t)*in_frag_length);
  l_new_frag_p->m_first_orig_char_index = in_first_orig_char_index;
  l_new_frag_p->m_orig_frag_length = in_orig_frag_length;
  if (in_orig_frag_length > 0) {
    l_new_frag_p->m_forward_xrefs_arr =
      in_rc_p->m_malloc_func(sizeof(size_t)*in_orig_frag_length);
  }
  else {
    l_new_frag_p->m_forward_xrefs_arr = NULL;
  }
  /* The caller is responsible for initializing the arrays
  ** m_frag_str, m_back_xrefs_arr and m_forward_xrefs_arr,
  ** as well as the following two fields. */
  l_new_frag_p->m_frag_state_free_func = NULL;
  l_new_frag_p->m_frag_state_p = NULL;

  /* Check if any attempt to allocate memory failed */
  if ((NULL == l_new_frag_p->m_frag_str)
      || (NULL == l_new_frag_p->m_back_xrefs_arr)
      || ((in_orig_frag_length > 0)
	  && (NULL == l_new_frag_p->m_forward_xrefs_arr))) {
    /* We ran out of memory, so clean up and return NULL. */
    if (NULL != l_new_frag_p->m_frag_str) {
      in_rc_p->m_free_func(l_new_frag_p->m_frag_str);
    }
    if (NULL != l_new_frag_p->m_back_xrefs_arr) {
      in_rc_p->m_free_func(l_new_frag_p->m_back_xrefs_arr);
    }
    if (NULL != l_new_frag_p->m_forward_xrefs_arr) {
      in_rc_p->m_free_func(l_new_frag_p->m_forward_xrefs_arr);
    }
    in_rc_p->m_free_func(l_new_frag_p);
    return(NULL);
  }

  /* All is OK, so return the pointer to the newly-allocated fragment.
  ** The arrays pointed at by this structure are uninitialized - it is
  ** expected that the caller will fill them completely with new
  ** values. */
  return(l_new_frag_p);
}

/*****************************************************************************
** Line constructor
** NULL is returned if memory allocation attempt failed.
*****************************************************************************/
SRenderedLine *
construct_line(const SRenderingContext *in_rc_p,
	       size_t in_screen_row_width)
{
  SRenderedLine *l_new_line_p = NULL;
  size_t l_index;                 /* Loop index */

  assert(NULL != in_rc_p);
  assert(validate_rendering_context(in_rc_p));
  assert(in_screen_row_width > 0);

  l_new_line_p =
    (SRenderedLine *) (in_rc_p->m_malloc_func(sizeof(SRenderedLine)));
  if (NULL == l_new_line_p) return(NULL);

  /* Set fields in the new line structure. */
  l_new_line_p->m_line_free_func = in_rc_p->m_free_func;
  l_new_line_p->m_screen_row_width = in_screen_row_width;
  l_new_line_p->m_total_frag_array_size =
    in_rc_p->m_initial_total_frag_array_size;
  l_new_line_p->m_used_frag_array_size = 0;
  l_new_line_p->m_frag_p_arr = in_rc_p->
    m_malloc_func(sizeof(SRenderedFragment *)
		  * in_rc_p->m_initial_total_frag_array_size);

  /* Check if any attempt to allocate memory failed */
  if (NULL == l_new_line_p->m_frag_p_arr) {
    /* We ran out of memory, so clean up and return NULL. */
    in_rc_p->m_free_func(l_new_line_p);
    return(NULL);
  }

  /* Initialize the m_frag_p_arr. */
  for (l_index = 0;
       l_index < in_rc_p->m_initial_total_frag_array_size;
       l_index++) {
    l_new_line_p->m_frag_p_arr[l_index] = NULL;
  }

  /* All is OK now.  The line structure is fully initialized.  The caller
  ** needs only to add to it at least one fragment. */
  return(l_new_line_p);
}

/*****************************************************************************
** Append a new fragment to the end of a line
*****************************************************************************/

E_RENDER_RETURN_CODES
line_append_fragment(const SRenderingContext *in_rc_p,
		     SRenderedLine *in_out_line_p,
		     SRenderedFragment *in_new_frag_p)
{
  assert(NULL != in_rc_p);
  assert(validate_rendering_context(in_rc_p));
  assert(NULL != in_out_line_p);
  assert(NULL != in_new_frag_p);

  /* Validate the SRenderedLine structure contents. */
  assert(NULL != in_out_line_p->m_line_free_func);
  assert(in_out_line_p->m_screen_row_width > 0);
  assert(in_out_line_p->m_total_frag_array_size > 0);
  assert(in_out_line_p->m_total_frag_array_size
	 >= in_out_line_p->m_used_frag_array_size);
  assert(NULL != in_out_line_p->m_frag_p_arr);
  /* We'll validate the contents of the array itself only if we'll
  ** copy it anyway, due to the need to enlarge it. */

  /* Enlarge m_frag_p_arr if necessary. */
  if (in_out_line_p->m_total_frag_array_size
      == in_out_line_p->m_used_frag_array_size) {
    /* Ugh, need to enlarge the array.
    ** We'll enlarge it to double its former size. */

    size_t l_new_total_size;
    SRenderedFragment **l_new_frag_p_arr;
    size_t l_index;

    l_new_total_size = 2 * in_out_line_p->m_total_frag_array_size;
    l_new_frag_p_arr = (SRenderedFragment **) in_rc_p->
      m_malloc_func(sizeof(SRenderedFragment *) * l_new_total_size);

    if (NULL == l_new_frag_p_arr) {
      /* Memory allocation failure */
      return(RENDER_FAILURE_NO_MEMORY);
    }

    /* Copy over data (validating it on the fly). */
    for (l_index = 0;
	 l_index < in_out_line_p->m_used_frag_array_size;
	 l_index++) {
      assert(NULL != in_out_line_p->m_frag_p_arr[l_index]);
      l_new_frag_p_arr[l_index] = in_out_line_p->m_frag_p_arr[l_index];
    }
    /* Clear the rest of the array */
    assert(l_index == in_out_line_p->m_used_frag_array_size);
    for (; l_index < l_new_total_size; l_index++) {
      l_new_frag_p_arr[l_index] = NULL;
    }

    /* Now modify *in_out_line_p as needed. */
    in_out_line_p->m_total_frag_array_size = l_new_total_size;
    assert(in_rc_p->m_free_func == in_out_line_p->m_line_free_func);
    in_out_line_p->m_line_free_func(in_out_line_p->m_frag_p_arr);
    in_out_line_p->m_frag_p_arr = l_new_frag_p_arr;
  }

  /* At this point, we are guaranteed to have room for an additional
  ** fragment. */
  assert(in_out_line_p->m_total_frag_array_size
	 > in_out_line_p->m_used_frag_array_size);

  in_out_line_p->m_frag_p_arr[in_out_line_p->m_used_frag_array_size]
    = in_new_frag_p;
  (in_out_line_p->m_used_frag_array_size)++;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Delete all fragments from a line, starting at a certain fragment
*****************************************************************************/

E_RENDER_RETURN_CODES
line_delete_fragments(const SRenderingContext *in_rc_p,
		      SRenderedLine *in_out_line_p,
		      size_t in_first_frag_to_free)
				  /* Index of first fragment to be freed
				  ** (the index of 1st fragment is 0). */
{
  size_t l_index;

  assert(NULL != in_rc_p);
  assert(validate_rendering_context(in_rc_p));
  assert(NULL != in_out_line_p);
  assert(in_first_frag_to_free < in_out_line_p->m_used_frag_array_size);

  /* Validate the SRenderedLine structure contents. */
  assert(NULL != in_out_line_p->m_line_free_func);
  assert(in_out_line_p->m_screen_row_width > 0);
  assert(in_out_line_p->m_total_frag_array_size > 0);
  assert(in_out_line_p->m_total_frag_array_size
	 >= in_out_line_p->m_used_frag_array_size);
  assert(NULL != in_out_line_p->m_frag_p_arr);

  /* Now, actually delete the fragments. */
  for (l_index = in_first_frag_to_free;
       l_index < in_out_line_p->m_used_frag_array_size;
       l_index++) {
    assert(NULL != in_out_line_p->m_frag_p_arr[l_index]);
    destroy_fragment((void *) in_out_line_p->m_frag_p_arr[l_index]);
    in_out_line_p->m_frag_p_arr[l_index] = NULL;
  }
  in_out_line_p->m_used_frag_array_size = in_first_frag_to_free - 1;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Append a new line to the end of SRenderedText
*****************************************************************************/

E_RENDER_RETURN_CODES
text_append_line(SRenderedText *in_out_rt_p,
				  /* Already contains pointer to the
				  ** current rendering context. */
		 SRenderedLine *in_new_line_p)
{
  assert(NULL != in_out_rt_p);
  assert(NULL != in_new_line_p);

  /* Validate the SRenderedText structure contents. */
  assert(NULL != in_out_rt_p->m_text_free_func);
  assert(in_out_rt_p->m_total_line_array_size > 0);
  assert(in_out_rt_p->m_total_line_array_size
	 >= in_out_rt_p->m_used_line_array_size);
  assert(NULL != in_out_rt_p->m_line_p_arr);
  /* We'll validate the contents of the array itself only if we'll
  ** copy it anyway, due to the need to enlarge it. */
  assert(NULL != in_out_rt_p->m_rendering_context_p);
  assert(validate_rendering_context(in_out_rt_p->m_rendering_context_p));
  /* We are not validating the contents of *in_new_line_p here. */

  /* Enlarge m_line_p_arr if necessary. */
  if (in_out_rt_p->m_total_line_array_size
      == in_out_rt_p->m_used_line_array_size) {
    /* Ugh, need to enlarge the array.
    ** We'll enlarge it to double its former size. */

    size_t l_new_total_size;
    SRenderedLine **l_new_line_p_arr;
    size_t l_index;

    l_new_total_size = 2 * in_out_rt_p->m_total_line_array_size;
    l_new_line_p_arr = (SRenderedLine **) in_out_rt_p->m_rendering_context_p->
      m_malloc_func(sizeof(SRenderedLine *) * l_new_total_size);

    if (NULL == l_new_line_p_arr) {
      /* Memory allocation failure */
      return(RENDER_FAILURE_NO_MEMORY);
    }

    /* Copy over data (validating it on the fly). */
    for (l_index = 0;
	 l_index < in_out_rt_p->m_used_line_array_size;
	 l_index++) {
      assert(NULL != in_out_rt_p->m_line_p_arr[l_index]);
      l_new_line_p_arr[l_index] = in_out_rt_p->m_line_p_arr[l_index];
    }
    /* Clear the rest of the array */
    assert(l_index == in_out_rt_p->m_used_line_array_size);
    for (; l_index < l_new_total_size; l_index++) {
      l_new_line_p_arr[l_index] = NULL;
    }

    /* Now modify *in_out_rt_p as needed. */
    in_out_rt_p->m_total_line_array_size = l_new_total_size;
    in_out_rt_p->m_rendering_context_p->m_free_func(in_out_rt_p->m_line_p_arr);
    in_out_rt_p->m_line_p_arr = l_new_line_p_arr;
  }

  /* At this point, we are guaranteed to have room for an additional
  ** line. */
  assert(in_out_rt_p->m_total_line_array_size
	 > in_out_rt_p->m_used_line_array_size);

  in_out_rt_p->m_line_p_arr[in_out_rt_p->m_used_line_array_size]
    = in_new_line_p;
  (in_out_rt_p->m_used_line_array_size)++;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Delete all lines from a SRenderedText, starting at a certain line.
*****************************************************************************/

E_RENDER_RETURN_CODES
text_delete_lines(SRenderedText *in_out_rt_p,
				  /* Already contains pointer to the
				  ** current rendering context. */
		  size_t in_first_line_to_free)
				  /* Index of first line to be freed
				  ** (the index of 1st line is 0). */
{
  size_t l_index;

  assert(NULL != in_out_rt_p);
  /* Validate the SRenderedText structure contents. */
  assert(NULL != in_out_rt_p->m_text_free_func);
  assert(in_out_rt_p->m_total_line_array_size > 0);
  assert(in_out_rt_p->m_total_line_array_size
	 >= in_out_rt_p->m_used_line_array_size);
  assert(NULL != in_out_rt_p->m_line_p_arr);
  assert(NULL != in_out_rt_p->m_rendering_context_p);
  assert(validate_rendering_context(in_out_rt_p->m_rendering_context_p));

  assert(in_first_line_to_free < in_out_rt_p->m_used_line_array_size);

  /* Now, actually delete the lines. */
  for (l_index = in_first_line_to_free;
       l_index < in_out_rt_p->m_used_line_array_size;
       l_index++) {
    assert(NULL != in_out_rt_p->m_line_p_arr[l_index]);
    destroy_line((void *) in_out_rt_p->m_line_p_arr[l_index]);
    in_out_rt_p->m_line_p_arr[l_index] = NULL;
  }
  in_out_rt_p->m_used_line_array_size = in_first_line_to_free - 1;
  return(RENDER_SUCCESS);
}

/*****************************************************************************
** Determine the offset in original text, which corresponds to 1st glyph
** in a line - for a line, which does not exist yet.
** Information in the previous line is used for this purpose.
*****************************************************************************/

E_RENDER_RETURN_CODES
locate_origtext_offset(const SRenderedText *in_rt_p,
				  /* Already contains pointer to the
				  ** current rendering context. */
		       size_t in_line,
				  /* Index of line for which we want the
				  ** offset in the original text. */
		       size_t *out_offset_p)
                                  /* Return value - contains the desired
				  ** offset.  If the previous line does
				  ** not exist, an error return code is
				  ** returned. */
{
  const SRenderedLine *l_line_p;
  const SRenderedFragment *l_frag_p;

  assert(NULL != in_rt_p);
  assert(NULL != out_offset_p);
  /* Validate the SRenderedText structure contents. */
  assert(in_rt_p->m_total_line_array_size > 0);
  assert(in_rt_p->m_total_line_array_size
	 >= in_rt_p->m_used_line_array_size);
  assert(NULL != in_rt_p->m_line_p_arr);
  assert(NULL != in_rt_p->m_rendering_context_p);
  assert(validate_rendering_context(in_rt_p->m_rendering_context_p));

  if (0 == in_line) {
    *out_offset_p = 0;
    return(RENDER_SUCCESS);
  }

  if (in_line > in_rt_p->m_used_line_array_size) {
    /* The previous line does not exist, so we can't return
    ** any answer. */
    return(RENDER_FAILURE_NOT_RENDERED_YET);
  }

  l_line_p = in_rt_p->m_line_p_arr[in_line-1];
  /* Validate the previous line's data structure */
  assert(NULL != l_line_p);
  assert(l_line_p->m_total_frag_array_size > 0);
  assert(l_line_p->m_used_frag_array_size > 0);
  assert(l_line_p->m_total_frag_array_size
	 >= l_line_p->m_used_frag_array_size);
  assert(NULL != l_line_p->m_frag_p_arr);

  l_frag_p = l_line_p->m_frag_p_arr[l_line_p->m_used_frag_array_size-1];
  /* Validate the previous line's last fragment pointer */
  assert(NULL != l_frag_p);

  *out_offset_p = l_frag_p->m_first_orig_char_index
    + l_frag_p->m_orig_frag_length;
  return(RENDER_SUCCESS);
}

/****************************************************************************/

/* End of plugin0.c */
