/*****************************************************************************
**
** structs0.h
** $Header: /home/omer/work/cvs_repository/Hebrew/include/structs0.h,v 1.7 1999/02/06 16:49:27 omer Exp $
**
** Declarations of constants, data types and structures used by Level 0 I18N
** text rendering plugin interface specifications.
**
** Copyright (C) 1999  Omer Zak
**
**
** This library is free software; you can redistribute it and/or
** modify it under the terms of the GNU Library General Public
** License as published by the Free Software Foundation; either
** version 2 of the License, or (at your option) any later version.
**
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
** Library General Public License for more details.
**
** You should have received a copy of the GNU Library General Public
** License along with this library; if not, write to the 
** Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
** Boston, MA  02111-1307  USA.
**
******************************************************************************
**
** Bug reports and fixes - to omerz@actcom.co.il
** Note:  by sending unsolicited commercial/political/religious/MailPush
**        E-mail messages (known also as "spam") to any E-mail address
**        mentioned in this file, you irrevocably agree to pay the
**        receipient US$500.- (plus any legal expenses incurred while
**        trying to collect the amount due) per unsolicited
**        commercial/political/religious/MailPush E-mail message - for
**        the service of receiving your E-mail message.
**
*****************************************************************************/

#ifndef _STRUCTS0_H_
#define _STRUCTS0_H_

#include <stdlib.h>

/*****************************************************************************
** Basic constants and data types                                           **
*****************************************************************************/

#ifndef BOOL
#define BOOL int
#endif /* BOOL */
#ifndef TRUE
#define TRUE 1
#endif /* TRUE */
#ifndef FALSE
#define FALSE 0
#endif /* FALSE */
#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif /* min */
#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif /* max */

/* The following two macro definitions were made conditional in order
** to allow their reconfiguration from a makefile. */
#ifndef PLUGIN_PATH_ENVIRONMENT_VARIABLE_NAME
#define PLUGIN_PATH_ENVIRONMENT_VARIABLE_NAME "TEXT_RENDERING_PLUGIN_PATH"
#endif /* PLUGIN_PATH_ENVIRONMENT_VARIABLE_NAME */

#ifndef HARDWIRED_PLUGIN_PATH_DEFAULT
#define HARDWIRED_PLUGIN_PATH_DEFAULT "/usr/lib/plugins/textrendering"
#endif /* HARDWIRED_PLUGIN_PATH_DEFAULT */
/* Some path for the above should be part of the FSSTND specification.
** E-mail message was sent to <Daniel.Quinlan@linux.org> and I am still
** waiting for answer from him.
**
** According to my proposal, the default path to configuration files
** for your plugin (assuming that its basename is your-plugin-name)
** should be:  /usr/lib/plugins/textrendering/your-plugin-name
*/

#define LEVEL0_VERSION_STRING "level0.0"
/* Used for validating the version of the SRenderingFunctions
** function pointers structure, which is used in the plugin.
** This string must be defined in all modules, which are used
** in level 0 plugins. */

#ifndef MALLOC_FREE_PROTOTYPES_DEFINED
typedef void *(MALLOC)(size_t in_size);  /* Prototype for caller
					 ** provided memory allocator */
typedef void (FREE)(void *in_p);         /* Prototype for caller
					 ** provided memory freer */
#define MALLOC_FREE_PROTOTYPES_DEFINED
#endif /* MALLOC_FREE_PROTOTYPES_DEFINED */

/*****************************************************************************
** Standard return values                                                   **
*****************************************************************************/
enum E_RENDER_RETURN_CODES {
  RENDER_SUCCESS = 0,
  RENDER_BEYOND_END_OF_TEXT,
  RENDER_FAILURE_GENERAL,
  RENDER_FAILURE_INVALID_ARGUMENTS,
  RENDER_FAILURE_NO_MEMORY,
  RENDER_FAILURE_NOT_RENDERED_YET
};

typedef enum E_RENDER_RETURN_CODES E_RENDER_RETURN_CODES;

/*****************************************************************************
** Data structure containing pointers to plugin functions exported to       **
** the client software.                                                     **
** Client software will never need to directly access the structure         **
** contents because it will call those functions only via suitable          **
** wrappers.                                                                **
**                                                                          **
** This data structure is used as interface to be used (indirectly)         **
** by the client software (such as text editors or text widgets).           **
** The basic idea is that the client needs to probe the plugin only for a   **
** single entry point (return_function_pointers()).                         **
** When called, this entry point returns pointer to a structure, which      **
** includes pointers to all functions, which the client software needs to   **
** invoke.                                                                  **
*****************************************************************************/

/* The function pointers below are typed as void*, in order to avoid having
** to declare all of them in this header file.
** Code, which invokes those function pointers, will take care of coercing
** them to the appropriate data type.  This is transparent to client
** software.
*/
typedef struct SRenderingFunctions {
  size_t m_struct_length;  /* set to sizeof(SRenderingFunctions) */
  char m_preamble_str[9];  /* set to "level0.0" */
  void *m_parse_config_func;
  void *m_delete_config_func;
  void *m_notify_modification_func;
  void *m_retrieve_rendered_line_func;
  char m_posteamble_str[9];  /* set to "level0.0" */
} SRenderingFunctions;


/*****************************************************************************
** Plugin descriptor structure                                              **
*****************************************************************************/

typedef struct SPlugin {
  void *m_lib_handle;              /* Handle for the loaded shared library */
  const SRenderingFunctions *m_funcptrs_p;
                                   /* Pointers to plugin functions */
} SPlugin;
/* When no plugin is bound, both fields in the above structure are NULL. */

/*****************************************************************************
** Context structure - packs together pointers and parameters, which are    **
** needed by several procedures.                                            **
*****************************************************************************/

typedef struct SRenderingContext {
  SPlugin *m_plugin_p;             /* Points at the plugin with which this
				   ** rendering context instance is bound.
				   ** To change a plugin, a new rendering
				   ** context instance has to be created. */
  void *m_plugin_config_p;         /* Opaque pointer to internal plugin
				   ** configuration information.
				   ** This pointer should have been made
				   ** available from the code which loads
				   ** and initializes the plugin.
				   */
  MALLOC *m_malloc_func;           /* Function for allocating
				   ** memory for output data
				   ** structures.
				   ** malloc() will be used if the
				   ** argument is NULL. */
  FREE *m_free_func;               /* Function for freeing
				   ** memory allocated by
				   ** m_malloc_func().
				   ** free() will be used if
				   ** the argument is NULL. */
  size_t m_bytes_per_glyph;        /* No. of bytes per glyph
				   ** of the rendered line.
				   ** Normally 1 or 2.
				   ** If the render_string()
				   ** instance being used does
				   ** not understand two-byte
				   ** glyphs, it just sets the
				   ** MSB to 0. */
  size_t m_initial_total_frag_array_size;
                                   /* Initial total size of *m_frag_p_arr
				   ** to be used when constructing a new
				   ** SRenderedLine instance. */
} SRenderingContext;
/* Note:  the default values are applied by set_clone_rendering_context().
** The functions, which use it, expect real values in the fields.
*/

/*****************************************************************************
** Data structure representing rendered glyphs.
**
** There are three data types, which are used in this data structure:
** SRenderedText - contains a pointer to an array of pointers to the data
**    structures representing the rendered lines, along with some auxiliary
**    variables.
** SRenderedLine - contains a pointer to an array of pointers to line
**    fragments, along with some auxiliary variables.
** SRenderedFragment - contains a pointer to a string containing the
**    glyphs of a line fragment, few attribute fields and some other
**    auxiliary information.
**
** Philosophical notes:
** 1. It is intended that a fragment will correspond to text fragments
**    commonly used by text editors.
** 2. All the characters in an original text fragment, and all the glyphs
**    in a rendered text fragment are expected to have the same attributes
**    (such as font, bold/underline/italic, LTR/RTL).
** 3. Fragments are grouped into lines according to the dictates of any
**    applicable line-breaking algorithm (which may of course break a
**    fragment into two if the fragment does not fit into a line).
** 4. When the client software asks for the rendered text from N lines,
**    the text included in N SRenderedLine structures will be returned
**    to the client.
** 5. The data structures will be opaque to the client (but not to the
**    plugin).  Any data accesses or manipulations shall be carried out
**    by means of procedures provided by the plugin.
*****************************************************************************/

typedef struct SRenderedFragment {
  FREE *m_frag_free_func;         /* Function to be used for freeing
				  ** dynamic memory areas pointed at
				  ** by this structure (with the exception
				  ** of m_frag_state_p, which has its own
				  ** free function). */

  /* Data, which will ultimately be used by the client */
  size_t m_frag_length;           /* Fragment length (in glyphs).
				  ** This variable defines the sizes of
				  ** the arrays pointed at by m_fragment_str
				  ** and m_xrefs_arr. */
  long m_orig_frag_attributes;    /* Fragment attributes inherited from
				  ** the original text - such as BOLD
				  ** or UNDERLINE */
  long m_added_frag_attributes;   /* Fragment attributes added by the
				  ** plugin - such as LTR/RTL. */
  void *m_frag_str;               /* The fragment data itself, not
				  ** necessarily null-terminated. */

  /* Data used to facilitate cursor and mouse position translations,
  ** as well as determination of the appropriate cursor shape to be used,
  ** when it is supposed to depend upon rendered text attributes. */
  size_t *m_back_xrefs_arr;       /* Array of cross reference pointers,
				  ** which allow the plugin to locate an
				  ** original text character, which is
				  ** associated with each rendered glyph.
				  ** The size of the array is given by
				  ** m_frag_length.
				  ** The values in this array are expressed
				  ** as offsets relative to 1st character
				  ** in the original text (whose position
				  ** is 0). */
  size_t m_first_orig_char_index; /* Offset of the first original text
				  ** character which was rendered in this
				  ** line fragment. */
  size_t m_orig_frag_length;      /* Length of the original text fragment
				  ** from which this rendered text fragment
				  ** was created. */
  size_t *m_forward_xrefs_arr;    /* Array of cross reference pointers,
				  ** which allow the plugin to locate
				  ** a rendered glyph, which is associated
				  ** with each original text character.
				  ** The size of the array is given
				  ** by m_orig_text_length.
				  ** The values in this array are expressed
				  ** as offsets relative to 1st glyph of the
				  ** current fragment (whose position
				  ** is 0).*/
  /* size_t m_beyond_last_orig_char_index;
  **     This variable, intended to be one plus the offset of the last
  **     original text character, which was rendered in this line fragment,
  **     is not needed because it is the same as
  **                   m_first_orig_char_index+m_orig_frag_length.
  */

  /* Internal state data, to be maintained by the plugin for its own
  ** purposes.  It is anticipated that the plugin will use some state
  ** data to optimize incremental work - spend as little work as
  ** possible re-rendering text, in response to minor modifications in
  ** the original text. */
  FREE *m_frag_state_free_func;   /* Function to be used for freeing
				  ** any memory area pointed at by
				  ** m_frag_state_p.
				  ** If it is NULL, this means that
				  ** m_frag_state_p does not point at
				  ** any dynamic memory structure. */
  void *m_frag_state_p;           /* Internal state associated with the
				  ** fragment - described by opaque
				  ** data structure.
				  ** An example would be to use it for
				  ** saving the contents of bidirectional
				  ** embedding stack according to Unicode
				  ** BiDi algorithms. */
} SRenderedFragment;

typedef struct SRenderedLine {
  FREE *m_line_free_func;         /* Function to be used for freeing
				  ** the *m_frag_p_arr array and the
				  ** SRenderedFragment items pointed
				  ** at by the array. */

  size_t m_screen_row_width;      /* Width of the screen row in glyphs.
				  ** Note:  in level 1 interface, this
				  ** will be expressed in terms of
				  ** pixels. */

  size_t m_total_frag_array_size; /* Total size of *m_frag_p_arr */
  size_t m_used_frag_array_size;  /* Actual number of fragments associated
				  ** with this line. */
  SRenderedFragment **m_frag_p_arr;
                                  /* Pointer to array-of-pointers
				  ** to rendered line fragments. */
} SRenderedLine;

typedef struct SRenderedText {
  FREE *m_text_free_func;         /* Function to be used for freeing
				  ** the *m_line_p_arr array and the
				  ** SRenderedLine items pointed
				  ** at by the array. */

  size_t m_total_line_array_size; /* Total size of *m_line_p_arr */
  size_t m_used_line_array_size;  /* Actual number of lines associated
				  ** with this text data structure. */
  SRenderedLine **m_line_p_arr;   /* Pointer to array-of-pointers
				  ** to rendered lines. */

  /* Rendering context with which this structure is bound */
  const SRenderingContext *m_rendering_context_p;
                                  /* Points at the rendering context
				  ** to which this SRenderedText instance
				  ** is bound.
				  ** It is not allowed to modify the
				  ** contents of the SRenderingContext
				  ** instance associated with an
				  ** SRenderedText instance.
				  ** But it is possible to clone a new
				  ** SRenderedText instance with a
				  ** different SRenderingContext. */
} SRenderedText;

#endif /* _STRUCT0_H_ */

/* End of structs0.h */
